/////////////////////////////////////////////////////////////////////////////
///
/// @file Facade.cpp
///
/// VM authentication interface.
///
/// @author shrike
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2020 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#include <errno.h>
#include <map>
#include <boost/foreach.hpp>
#include <prlcommon/Logging/Logging.h>

#include "Facade.h"

Prl::Expected<FILE*, int> Facade::download(const char* path_) const
{
	char* n = tmpnam(NULL);
	if (NULL == n)
	{
		WRITE_TRACE(DBG_FATAL, "tmpnam failed");
		return -4;
	}
	if (-1 == guestfs_download(m_handle, path_, n))
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_download: %s", guestfs_last_error(m_handle));
		return -1;
	}
	FILE* output = fopen(n, "r");
	if (NULL == output)
	{
		WRITE_TRACE(DBG_FATAL, "fopen(\"%s\"): %d", path_, errno);
		unlink(n);
		return -4;
	}
	unlink(n);
	return output;
}

void Facade::free(char** list_)
{
	if (NULL == list_)
		return;

	for (char** p = list_; *p != NULL; ++p)
	{
		::free(*p);
	}
	::free(list_);
}

Prl::Expected<void, int> Facade::mount(const std::string& root_) const
{
	char** p = guestfs_inspect_get_mountpoints(m_handle, root_.data());
	if (NULL == p)
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_inspect_get_mountpoints: %s", guestfs_last_error(m_handle));
		return -1;
	}
	typedef std::map<std::string, std::string> map_type;
	map_type m;
	for (char** i = p; *i != NULL; i += 2)
	{
		m[*i] = *(i + 1);
	}
	free(p);
	BOOST_FOREACH(map_type::const_reference x, m)
	{
		if (0 != guestfs_mount_ro(m_handle, x.second.data(), x.first.data()))
		{
			WRITE_TRACE(DBG_FATAL, "guestfs_mount_ro: %s",
				guestfs_last_error(m_handle));
		}
	}
	return Prl::Expected<void, int>();
}

Prl::Expected<std::string, int> Facade::getOsRoot() const
{
	char** r = guestfs_inspect_os(m_handle);
	if (NULL == r)
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_inspect_os: %s", guestfs_last_error(m_handle));
		return -1;
	}
	if (NULL == r[0])
	{
		free(r);
		WRITE_TRACE(DBG_FATAL, "no operating system found");
		return -2;
	}
	std::string output = r[0];
	free(r);
	return output;
}

Prl::Expected<Facade::os_type, int> Facade::getOsType(const std::string& root_) const
{
	char* t = guestfs_inspect_get_type(m_handle, root_.data());
	if (NULL == t)
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_inspect_get_type: %s", guestfs_last_error(m_handle));
		return -1;
	}
	if (0 == strcmp(t, "linux"))
	{
		::free(t);
		return os_type(Linux(m_handle));
	}
	if (0 == strcmp(t, "windows"))
	{
		::free(t);
		t = guestfs_inspect_get_windows_systemroot(m_handle, root_.data());
		if (NULL == t)
		{
			WRITE_TRACE(DBG_FATAL, "guestfs_inspect_get_windows_systemroot: %s",
				guestfs_last_error(m_handle));
			return -1;
		}
		std::string r = t;
		::free(t);
		return os_type(Windows::Engine(m_handle, r));
	}
	::free(t);
	WRITE_TRACE(DBG_FATAL, "unsupported operating system type");
	return -3;
}
