/////////////////////////////////////////////////////////////////////////////
///
/// @file IHive.h
///
/// This module describes some functions which may work with registry hive
/// as data storage: read from file, keep in memory, write to file etc.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#ifndef	__HIVE__INTERFACE__
#define	__HIVE__INTERFACE__

#include <QString>
#include "WinRegistryErrors.h"

///////////////////////////////////////////////////////////////////////////////
// class IHive
// This interface implements some abstract storage which may
// read data from hive, keep these data in memory and write them
// back. Besides these functions it may detect type of hive.

class IHive
{
protected:
	virtual ~IHive() {};
public:
	virtual err_status_t Open(const QString& sHiveName) = 0;
	virtual err_status_t Open(FILE* hive_) = 0;
	virtual err_status_t Write() = 0;
	virtual err_status_t Close() = 0;
	virtual QByteArray* GetStorage() = 0;
	virtual void Free() = 0;
	virtual void SetDirtySign(bool bDirty = true) = 0;


	/**
	 * Types of Windows hives.
	 *
	 * @author igorl
	 */
	enum HiveVersion
	{
		HV_UNKNOWN = 0,
		HV_WIN9X,
		HV_WINNT
	};


	/**
	 * Detects type of specified hive based on first data of this hive.
	 * Now first four bytes are enough.
	 *
	 * @author igorl
	 *
	 * @param[in] pData is the pointer to buffer containde first data
	 * of specified hive.
	 * @param[in] iSize is the size of buffer containde first data
	 * of specified hive.
	 *
	 * @return Type of specified hive.
	 */
	static HiveVersion GetHiveVersion(quint8* pData, int iSize)
	{
		if(!pData || iSize < (int) sizeof(unsigned))
			return  HV_UNKNOWN;

		unsigned*  pSignature = reinterpret_cast<unsigned*>(pData);

		switch(*pSignature)
		{
			/*
			// 0x47455243 = CREG
			// We don support registry of Win9X now.
			case 0x47455243: return  HV_WIN9X;
			*/
			// 0x66676572 = regf
			case 0x66676572: return  HV_WINNT;
			default        : return  HV_UNKNOWN;
		}

		return  HV_UNKNOWN;
	}
};

#endif // __HIVE__INTERFACE__
