/////////////////////////////////////////////////////////////////////////////
///
/// @file NtBasicFunctions.cpp
///
/// This module implements "low-level" functions for working with keys and
/// values of "NT" registry hive.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <QStringList>
#include <prlcommon/Logging/Logging.h>
#include "RegTypes.h"
#include "NtBasicFunctions.h"

#define HHS		quint32(HIVE_HDR_SIZE)
#define LFS		quint32(LEN_FIELD_SIZE)

static inline ptrdiff_t U32_2PTRDIFF(const quint32& x)
{
	Q_ASSERT_X((x) != quint32(-1), "Win registry basic functions", "Invalid pointer operation with quint32(-1)");
	return static_cast<ptrdiff_t>(x);
}

///////////////////////////////////////////////////////////////////////////////
// class CNtBasicFunctions

/**
 * Compares the first iSize bytes of pLeft and pRight without regard to
 * the case of letters in the two buffers. The function converts all
 * lowercase characters into uppercase (!) and then performs the comparison.
 *
 * @author igorl
 *
 * @param[in] pLeft is the pointer to first buffer.
 * @param[in] pRight is the pointer to second buffer.
 * @param[in] iSize is the number of characters.
 *
 * @return -1 if pLeft is less than pRight, 0 if pLeft is identical to pRight,
 * 1 if pLeft is greater than pRight.
 */
int CNtBasicFunctions::Compare(const char* pLeft,
									const char* pRight, int iSize) const
{
	// NB!
	// memicmp and strncasecmp work another way: they convert all uppercase
	// characters into lowercase and then performs compasion. Because of this
	// comparision of these strings
	// "IDE#CdRomATAPI_CD-ROM____________________________"
	// "IDE#CdRomATAPI_CD-ROM_DRIVE-20X__________________"
	// is incorrect from point of view of Windows registry.

	char  chDif;

	for(int i = 0; i < iSize; ++i)
	{
		chDif = toupper(*pLeft) - toupper(*pRight);

		if(chDif == 0)
		{
			++pLeft;
			++pRight;
		}
		else
		{
			if(chDif < 0)
				return -1;
			else
				return  1;
		}
	}

	return  0;
}


//
// Function looks for key among of child keys of given key.
//
// Parameters:
//    pNK [in]          - parent key.  Exactly among child keys of this key we will look for given subkey.
//    sSubKeyName [in]  - name of subkey. NB: now we examine only ASCII names.
//    pos, ri_pos [out] - these indexes determine position of found subkey in "indexes array" or are equal
//                        to -1 both if subkey was not found.
//
// Return value:
//    -1 if subkey was not found or its absolute offset in hive otherwise.
//
quint32 CNtBasicFunctions::FindSubKey(NK_CELL* pNK, const char* szSubKeyName, quint16& pos, quint16& ri_pos)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::FindSubKey]");
	*/

	quint8		*pData = NULL;
	quint16		IndexType;
	quint32		IndexOffset;
	RI_CELL		*pRI = NULL;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::FindSubKey] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.01:");

		return  (quint32) -1;
	}

	// Determine child index type
	IndexType = *((quint16*) (pData + U32_2PTRDIFF(HHS + pNK->OffSubKeysIndex + LFS)));

	if(IndexType == ID_RI)
	{
		pRI    = (RI_CELL*) (pData + U32_2PTRDIFF(HHS + pNK->OffSubKeysIndex + LFS));
		ri_pos = 0;
	}
	else
	{
		pRI    = NULL;
		ri_pos = (quint16) -1;
	}

	do
	{
		// Get offset child index
		if(pRI)
		{
			// In case of 'ri' index get next sub child index
			IndexOffset = HHS + pRI->Offset[ri_pos] + LFS;
			IndexType   = *((quint16*) (pData + U32_2PTRDIFF(IndexOffset)));
		}
		else
		{
			IndexOffset = HHS + pNK->OffSubKeysIndex + LFS;
		}

		if(IndexType == ID_LF || IndexType == ID_LH)
		{
			LF_LH_CELL	*pLF = (LF_LH_CELL*) (pData + U32_2PTRDIFF(IndexOffset));

			// Runing on subkeys
			pos = 0;
			while(pos < pLF->KeysCount)
			{
				NK_CELL		*pNKnext = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + (pLF->Hash[pos]).Offset + LFS));

				if(strlen(szSubKeyName) == pNKnext->NameLength)
				{
					// Lengthes of names are equal

					// Name of key is case insesitive (according to MSDN)
					// FIXME: UNICODE support
					if(!Compare(szSubKeyName, pNKnext->Name, pNKnext->NameLength))
					{
						// Names are equal too

						return  HHS + (pLF->Hash[pos]).Offset + LFS;
					}
				}

				// To next subkey
				pos++;
			}
		}

		if(IndexType == ID_LI)
		{
			LI_CELL		*pLI = (LI_CELL*) (pData + U32_2PTRDIFF(IndexOffset));

			// Runing on subkeys
			pos = 0;
			while(pos < pLI->KeysCount)
			{
				NK_CELL		*pNKnext = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + pLI->Offset[pos] + LFS));

				if(strlen(szSubKeyName) == pNKnext->NameLength)
				{
					// Lengthes of names are equal

					// Name of key is case insesitive (according to MSDN)
					// FIXME: UNICODE support
					if(!Compare(szSubKeyName, pNKnext->Name, pNKnext->NameLength))
					{
						// Names are equal too

						return  HHS + pLI->Offset[pos] + LFS;
					}
				}

				// To next key
				pos++;
			}
		}

		// To next sub child index
		if(pRI)
			ri_pos++;
	}
	while(pRI && ri_pos < pRI->IndexesCount);	// Loop on array of "indexes arrrayes" in case of RI

	// Subkey was not found
	pos    = (quint16) -1;
	ri_pos = (quint16) -1;

	return   (quint32) -1;
}

//
// Function opens existing key.
//
// Parameters:
//    sKeyName       [in]  - name of key. NB: now we examine only ASCII names.
//    dwKeyOffset    [out] - absolute offset of subkey with given name. If this subkey is not found this parameter
//                           will be equal -1.
//    dwParentOffset [in]  - absolute offset of parent key. Exactly in this key we will look for subkey with given
//                           name. If we want to look for subkey from root key we follow to set this offset  equal
//                           -1. This parameter is not obligatory, its default value is equal to -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::OpenKey(const QString& sKeyName, quint32& dwKeyOffset, quint32 dwParentOffset)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::OpenKey]");
	*/

	quint8		*pData = NULL;
	QStringList listNames = sKeyName.split("\\");
	quint32		dwNKoff;
	NK_CELL		*pNK = NULL;
	quint16		pos, ri_pos;	// these variables are not realy used in this function
	int			i;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::OpenKey] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.02:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	if(sKeyName.isEmpty())
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::OpenKey] Name of subkey is empty.");
		WRITE_TRACE(DBG_FATAL, "OA00002.03:");

		return  WinRegistry::REG_KEY_NOT_FOUND_ERROR;
	}

	if(dwParentOffset == (quint32) -1)
	{
		// Offset was not determine, lets go from root key

		dwNKoff = HHS + m_pRawData->GetRootCellOffset() + LFS;
	}
	else
		dwNKoff = dwParentOffset;

	for(i = 0; i < listNames.size(); i++)
	{
		// pNK addresses key among children of which we will look for subkey with
		// name listNames[i]
		pNK = (NK_CELL*) (pData + U32_2PTRDIFF(dwNKoff));

		if(pNK->Signature != ID_NK)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::OpenKey] Incorrect NK cell (i = %d, %s).", i, listNames.at(i).toAscii().data());
			WRITE_TRACE(DBG_FATAL, "OA00002.04:\t%d;\t%s", i, listNames.at(i).toAscii().data());

			return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
		}

		if(pNK->SubKeysCount > 0)
		{
			// FIXME: UNICODE support
			dwNKoff = FindSubKey(pNK, listNames.at(i).toAscii().data(), pos, ri_pos);

			if(dwNKoff == (quint32) -1)
				break;
		}
		else
		{
			dwNKoff = (quint32) -1;

			break;
		}
	}

	if(i == listNames.size() && dwNKoff > 0)
	{
		// We moved all names and search was successful, i.e. we found
		// given subkey

		dwKeyOffset = dwNKoff;

		return  PRL_ERR_SUCCESS;
	}

	dwKeyOffset = (quint32) -1;

	return  WinRegistry::REG_KEY_NOT_FOUND_ERROR;
}

//
// Function create new LF/LH "indexes array"
//
// Parameters:
//    szKeyName   [in]  - name of key. We will calculate hash value on base of this name. NB: now we examine
//                        only ASCII names.
//    NKnewOff    [in]  - absolute offset of cell (!) in which this key was placed key.
//    IndexOffset [out] - absolute offset of new cell (!) in which new "indexes array" will be placed. If some
//                        error happens it will be equal to -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::CreateLfLh(const char* szKeyName, quint32 NKnewOff, quint32& IndexOffset)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateLfLh]");
	*/

	quint8			*pData = NULL;
	LF_LH_CELL		*pLF = NULL;
	quint32			dwMinor;
	quint16			IndexType;
	err_status_t	Err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateLfLh] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.05:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	dwMinor = m_pRawData->GetMinorVersion();

	if(dwMinor != (quint32) -1)
	{
		// According to Paul Lee LF type of "indexes array" is existed till 4 version inclusive
		if(dwMinor <= 4)
			IndexType = ID_LF;
		else
			IndexType = ID_LH;
	}
	else
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateLfLh] Can not determine kind and(or) version of hive.");
		WRITE_TRACE(DBG_FATAL, "OA00002.06:");

		return  WinRegistry::REG_UNKNOWN_ERROR;
	}

	// Allocate space for new cell
	Err = m_pRawData->AllocCell(sizeof(LF_LH_CELL), IndexOffset);

	if(PRL_FAILED(Err))
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateLfLh] Can not allocate space (0x%x bytes) for new 'indexes array' (error code: 0x%x).", sizeof(LF_LH_CELL), Err);
		WRITE_TRACE(DBG_FATAL, "OA00002.07:\t%x;\t%x", sizeof(LF_LH_CELL), Err);

		return  Err;
	}

	pLF = (LF_LH_CELL*) (pData + U32_2PTRDIFF(IndexOffset + LFS));

	// Index cell type
	pLF->Signature          = IndexType;
	// Count of addressed keys
	pLF->KeysCount          = 1;
	// Offset of key itself
	((pLF->Hash)[0]).Offset = NKnewOff - HHS;

	//
	// Calculate key hash (FIXME: may be design separate function)
	//

	if(pLF->Signature == ID_LF)
	{
		// FIXME: what about UNICODE (according to Paul Lee it should be zero)

		size_t i;
		size_t iHashLen = sizeof(((pLF->Hash)[0]).Name);

		for(i = 0; i < iHashLen; i++)
		{
			if(i > strlen(szKeyName))
				((pLF->Hash)[0]).Name[i] = 0;
			else
				((pLF->Hash)[0]).Name[i] = szKeyName[i];
		}
	}
	else
	{
		if(pLF->Signature == ID_LH)
		{
			// FIXME: according to mityak@ for UNICODE name hash is calculated
			// by the same rule but under "words" (not "bytes").

			size_t		i;
			quint32		hash;

			hash = 0;
			for(i = 0; i < strlen(szKeyName); i++)
			{
				hash *= 37;
				hash += toupper(szKeyName[i]);
			}

			*((quint32*)((pLF->Hash)[0]).Name) = hash;
		}
		else
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateLfLh] Incorrect index cell (0x%x).", pLF->Signature);
			WRITE_TRACE(DBG_FATAL, "OA00002.08:\t%x", pLF->Signature);

			// Now it doesn't matter error handling
			m_pRawData->FreeCell(IndexOffset);

			return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
		}
	}

	return  PRL_ERR_SUCCESS;
}

//
// Function update existing LF/LH "indexes array"
//
// Parameters:
//    szKeyName      [in]  - name of key. We will calculate hash value on base of this name. NB: now we examine
//                           only ASCII names.
//    NKnewOff       [in]  - absolute offset of cell (!) in which this key was placed key.
//    IndexOffsetOld [in]  - relative offset of old "indexes array".  This array will be cleared out in case of
//                           its successful updating but value of this offset will not be changed (!).
//    IndexOffsetNew [out] - absolute offset of new cell (!) in which updated "indexes array" will be placed. If
//                           some error happens it will be equal to -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::UpdateLfLh(const char* szKeyName, quint32 NKnewOff, quint32 IndexOffsetOld, quint32& IndexOffsetNew)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[UpdateLfLh]");
	*/

	quint8			*pData = NULL;
	quint32			size;
	LF_LH_CELL		*pLFold = NULL, *pLF = NULL;
	int				nlen, cmp, pos;
	quint16			i, j;
	err_status_t	Err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLfLh] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.09:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pLFold = (LF_LH_CELL*) (pData + U32_2PTRDIFF(HHS + IndexOffsetOld + LFS));
	nlen   = strlen(szKeyName);

	if(pLFold->Signature != ID_LF && pLFold->Signature != ID_LH)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLfLh] Incorrect index cell type.");
		WRITE_TRACE(DBG_FATAL, "OA00002.10:");

		return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
	}

	if(pLFold->KeysCount >= MAX_LF_LH_COUNT)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLfLh] 'Indexes array' is full.");
		WRITE_TRACE(DBG_FATAL, "OA00002.11:");

		return  WinRegistry::REG_INDEXES_ARRAY_FULL_ERROR;
	}

	// Size of new "indexes array" (it contains additional new index)
	size = sizeof(LF_LH_CELL) + sizeof(LF_LH_CELL::LF_LH_HASH) * pLFold->KeysCount;

	// Allocate space for new "indexes array"
	Err = m_pRawData->AllocCell(size, IndexOffsetNew);

	if(PRL_FAILED(Err))
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLfLh] Can not allocate space (0x%x bytes) for 'indexes array' (error code: %d).", size, Err);
		WRITE_TRACE(DBG_FATAL, "OA00002.12:\t%x;\t%d", size, Err);

		return Err;
	}

	pLF = (LF_LH_CELL*) (pData + U32_2PTRDIFF(IndexOffsetNew + LFS));

	// Index cell type
	pLF->Signature = pLFold->Signature;
	// Count of addressed keys
	pLF->KeysCount = pLFold->KeysCount + 1;

	//
	// The following loop is meaned for
	// 1. Loking for position of new key in "indexes array". These "indexes" are arranged
	// by lexicographical order.
	// 2. Copying old "indexes array" to new one. Element for new key will be empty.
	//

	for(i = 0, j = 0, pos = -1; i < pLFold->KeysCount; i++)
	{
		// Looking for position of new key
		if(pos == -1)
		{
			NK_CELL*	pNK = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + ((pLFold->Hash)[i]).Offset + LFS));

			// FIXME: This comparision is case-insensitive. On Vista I have some
			// problem because of this.
			// May be already fixed (see UCMDB20485)?

			cmp = Compare(szKeyName, pNK->Name, (nlen > pNK->NameLength) ? pNK->NameLength : nlen);

			if(cmp == 0 && nlen == pNK->NameLength)
			{
				LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLfLh] Key '%s' already exists.", szKeyName);
				WRITE_TRACE(DBG_FATAL, "OA00002.13:\t%s", szKeyName);

				m_pRawData->FreeCell(IndexOffsetNew);

				return  WinRegistry::REG_KEY_ALREADY_EXISTS_ERROR;
			}

			if(cmp < 0 || (cmp == 0 && nlen < pNK->NameLength))
			{
				// Position was found

				pos = i;
				// Skip position of new key
				j++;
			}
		}

		// Copying element of old "indexes array"
		((pLF->Hash)[j]).Offset = ((pLFold->Hash)[i]).Offset;
		memcpy(((pLF->Hash)[j]).Name, ((pLFold->Hash)[i]).Name, sizeof(((pLFold->Hash)[i]).Name));

		j++;
	}

	if(pos < 0)
		pos = pLFold->KeysCount;

	// Offset of new key itself
	((pLF->Hash)[pos]).Offset = NKnewOff - HHS;

	//
	// Calculate key hash (FIXME: may be design separate function)
	//

	if(pLF->Signature == ID_LF)
	{
		// FIXME: what about UNICODE (according to Paul Lee it should be zero)

		unsigned	iHashLen = sizeof(((pLF->Hash)[pos]).Name);

		for(i = 0; i < iHashLen; i++)
		{
			if(i > strlen(szKeyName))
				((pLF->Hash)[pos]).Name[i] = 0;
			else
				((pLF->Hash)[pos]).Name[i] = szKeyName[i];
		}
	}
	else
	{
		if(pLF->Signature == ID_LH)
		{
			// FIXME: according to mityak@ for UNICODE name hash is calculated
			// by the same rule but under "words" (not "bytes").

			quint32		hash;

			hash = 0;
			for(i = 0; i < strlen(szKeyName); i++)
			{
				hash *= 37;
				hash += toupper(szKeyName[i]);
			}

			*((quint32*) ((pLF->Hash)[pos]).Name) = hash;
		}
		else
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLfLh] Incorrect index cell.");
			WRITE_TRACE(DBG_FATAL, "OA00002.14:");

			// Now it doesn't matter error handling
			m_pRawData->FreeCell(IndexOffsetNew);

			return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
		}
	}

	// Free old "indexes array"
	return  m_pRawData->FreeCell(IndexOffsetOld + HHS);
}

//
// Function update existing LI "indexes array"
//
// Parameters:
//    szKeyName      [in]  - name of key. We will calculate hash value on base of this name. NB: now we examine
//                           only ASCII names.
//    NKnewOff       [in]  - absolute offset of cell (!) in which this key was placed key.
//    IndexOffsetOld [in]  - relative offset of old "indexes array".  This array will be cleared out in case of
//                           its successful updating but value of this offset will not be changed (!).
//    IndexOffsetNew [out] - absolute offset of new cell (!) in which updated "indexes array" will be placed. If
//                           some error happens it will be equal to -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::UpdateLi(const char* szKeyName, quint32 NKnewOff, quint32 IndexOffsetOld, quint32& IndexOffsetNew)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[UpdateLi]");
	*/

	quint8			*pData = NULL;
	quint32			size;
	LI_CELL			*pLIold = NULL, *pLI = NULL;
	int				nlen, cmp, pos;
	quint16			i, j;
	err_status_t	Err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLi] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.15:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pLIold = (LI_CELL*) (pData + U32_2PTRDIFF(HHS + IndexOffsetOld + LFS));
	nlen   = strlen(szKeyName);

	if(pLIold->Signature != ID_LI)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLi] Incorrect index cell type.");
		WRITE_TRACE(DBG_FATAL, "OA00002.16:");

		return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
	}

	if(pLIold->KeysCount >= MAX_LI_COUNT)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLi] 'Indexes array' is full.");
		WRITE_TRACE(DBG_FATAL, "OA00002.17:");

		return  WinRegistry::REG_INDEXES_ARRAY_FULL_ERROR;
	}

	// Size of new "indexes array" (it contains additional new index)
	// P.S.
	// sizeof(quint32) - size of offset filed in LI_CELL struture
	size = sizeof(LI_CELL) + sizeof(quint32) * pLIold->KeysCount;

	// Allocate space for new "indexes array"
	Err = m_pRawData->AllocCell(size, IndexOffsetNew);

	if(PRL_FAILED(Err))
		return Err;

	pLI = (LI_CELL*) (pData + U32_2PTRDIFF(IndexOffsetNew + LFS));

	// Index cell type
	pLI->Signature = pLIold->Signature;
	// Count of addressed keys
	pLI->KeysCount = pLIold->KeysCount + 1;

	//
	// The following loop is meaned for
	// 1. Loking for position of new key in "indexes array". These "indexes" are arranged
	// by lexicographical order.
	// 2. Copying old "indexes array" to new one. Element for new key will be empty.
	//

	for(i = 0, j = 0, pos = -1; i < pLIold->KeysCount; i++)
	{
		// Looking for position of new key
		if(pos == -1)
		{
			NK_CELL	*pNK = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + (pLIold->Offset)[i] + LFS));

			// FIXME: This comparision is case-insensitive. On Vista I have some
			// problem because of this.
			// May be already fixed (see UCMDB20485)?

			cmp = Compare(szKeyName, pNK->Name, (nlen > pNK->NameLength) ? pNK->NameLength : nlen);

			if(cmp == 0 && nlen == pNK->NameLength)
			{
				LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateLi] Key '%s' already exists.", szKeyName);
				WRITE_TRACE(DBG_FATAL, "OA00002.18:\t%s", szKeyName);

				m_pRawData->FreeCell(IndexOffsetNew);

				return  WinRegistry::REG_KEY_ALREADY_EXISTS_ERROR;
			}

			if(cmp < 0 || (cmp == 0 && nlen < pNK->NameLength))
			{
				// Position was found

				pos = i;
				// Skip position of new key
				j++;
			}
		}

		// Copying element of old "indexes array"
		(pLI->Offset)[j] = (pLIold->Offset)[i];

		j++;
	}

	if(pos < 0)
		pos = pLIold->KeysCount;

	// Offset of key itself
	(pLI->Offset)[pos] = NKnewOff - HHS;

	// Free old "indexes array"
	return  m_pRawData->FreeCell(IndexOffsetOld + HHS);
}


/**
 * Adds new Lf/Lh index cell to existed Ri cell, i.e. calculates size of
 * new Ri cell, allocates needed space, fills new Ri cell and frees old
 * one.
 *
 * @author igorl
 *
 * @param[in] dwRiOffOld is relative offset of existed Ri cell.
 * @param[in] dwLfLhOff is absolute offset of adding Lf/Lh index cell.
 * @param[out] dwRiOffNew is absolute offset of new Ri cell.
 *
 * @return PRL_ERR_SUCCESS if there are no error or detailed error code
 * otherwise.
 */
err_status_t CNtBasicFunctions::AddLfLhToRi(
								quint32		dwRiOffOld,
								quint32		dwLfLhOff,
								quint32&	dwRiOffNew)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::AddLfLhToRi]");
	*/

	quint8*			pData  = NULL;
	quint32			size;
	RI_CELL*		pRIold = NULL;
	RI_CELL*		pRI    = NULL;
	LF_LH_CELL*		pLF    = NULL;
	err_status_t	err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::AddLfLhToRi]"
													" Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.73:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pRIold = reinterpret_cast<RI_CELL*>(pData + U32_2PTRDIFF(HHS + dwRiOffOld + LFS));

	if(pRIold->Signature != ID_RI)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::AddLfLhToRi]"
											" Index cell type is not RI.");
		WRITE_TRACE(DBG_FATAL, "OA00002.74:");

		return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
	}

	pLF = reinterpret_cast<LF_LH_CELL*>(pData + U32_2PTRDIFF(dwLfLhOff + LFS));

	if(pLF->Signature != ID_LF && pLF->Signature != ID_LH)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::AddLfLhToRi]"
										" Index cell type is not LF/LH.");
		WRITE_TRACE(DBG_FATAL, "OA00002.75:");

		return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
	}

	// Size of new "indexes array" (it contains additional new index)
	// P.S.
	// sizeof(quint32) - size of offset filed in RI_CELL struture
	size = sizeof(RI_CELL) + sizeof(quint32) * pRIold->IndexesCount;

	// Allocate space for new "indexes array"
	err  = m_pRawData->AllocCell(size, dwRiOffNew);

	if(PRL_FAILED(err))
		return err;

	pRI = reinterpret_cast<RI_CELL*>(pData + U32_2PTRDIFF(dwRiOffNew + LFS));

	// Index cell type
	pRI->Signature = pRIold->Signature;
	// Count of addressed keys
	pRI->IndexesCount = pRIold->IndexesCount + 1;

	// Copy offsets of existed indexes
	for(quint32 i = 0; i < pRIold->IndexesCount; i++)
		(pRI->Offset)[i] = (pRIold->Offset)[i];

	// Add offset of new Lf/Lh index
	(pRI->Offset)[pRI->IndexesCount - 1] = dwLfLhOff - HHS;

	return  m_pRawData->FreeCell(dwRiOffOld + HHS);
}


/**
 * We have full Li index cell (addressed by dwLIoff) and new adding key
 * (with specified name and offset) is hit to this Li cell. We should find
 * corresponded place for this new key, place it there, shift other keys
 * and receive name and offset of pushed old key.
 * NB!
 * This function checks fullness of Li index cell.
 *
 * @author igorl
 *
 * @param[in/out] sKeyName is the name of new adding key on input and
 * the name of pushed old key on output.
 * @param[in/out] dwNKoff is the offset of new adding key on input and
 * the offset of pushed old key on output. Both offsets are absolute.
 * @param[in] dwLIoff is the relative offset of editing Li cell.
 *
 * @return PRL_ERR_SUCCESS if there are no error or detailed error code
 * otherwise.
 */
err_status_t CNtBasicFunctions::Swap(
								QString&	sKeyName,
								quint32&	dwNKoff,
								quint32		dwLIoff)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::Swap]");
	*/

	quint8*			pData = NULL;
	// FIXME: Unicode support
	QByteArray      aName = sKeyName.toAscii();
	char*			pName = aName.data();
	LI_CELL*		pLI = NULL;
	NK_CELL*		pNK = NULL;
	int				nlen, cmp, pos;
	quint32			dwTempNKoff = static_cast<quint32>(-1);
	quint16			i;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::Swap]"
											" Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.76:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pLI = reinterpret_cast<LI_CELL*>(pData + U32_2PTRDIFF(HHS + dwLIoff + LFS));

	if(pLI->Signature != ID_LI)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::Swap]"
											" Index cell type is not LI.");
		WRITE_TRACE(DBG_FATAL, "OA00002.77:");

		return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
	}

	if(pLI->KeysCount < MAX_LI_COUNT)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::Swap]"
											" 'Indexes array' is not full.");
		WRITE_TRACE(DBG_FATAL, "OA00002.78:");

		// Not very good, but ...
		return  WinRegistry::REG_UNKNOWN_ERROR;
	}

	//
	// Looking for position of new key in "indexes array". These "indexes" are
	// arranged by lexicographical order.
	//

	nlen = strlen(pName);

	for(i = 0, pos = static_cast<quint16>(-1); i < pLI->KeysCount; i++)
	{
		pNK = reinterpret_cast<NK_CELL*>(pData + U32_2PTRDIFF(HHS + (pLI->Offset)[i] + LFS));

		// FIXME: This comparision is case-insensitive. On Vista I have some
		// problem because of this.
		// May be already fixed (see UCMDB20485)?

		cmp = Compare(pName, pNK->Name,
							(nlen > pNK->NameLength) ? pNK->NameLength : nlen);

		if(cmp == 0 && nlen == pNK->NameLength)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::Swap]"
								" Key '%s' already exists.", pName);
			WRITE_TRACE(DBG_FATAL, "OA00002.79:\t%s", pName);

			return  WinRegistry::REG_KEY_ALREADY_EXISTS_ERROR;
		}

		if(cmp < 0 || (cmp == 0 && nlen < pNK->NameLength))
		{
			// Position was found

			pos = i;
			break;
		}
	}

	// To be on the safe side :)
	// FIXME: Is this key will be found always?
	if(pos < 0)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::Swap]"
								" Key '%s' is not found.", pName);
		WRITE_TRACE(DBG_FATAL, "OA00002.80:\t%s", pName);

		return  WinRegistry::REG_KEY_NOT_FOUND_ERROR;
	}

	//
	// "Swap" itself: inserts offset of new key, shifts offsets of existed
	// keys, receives offset of key remained without "place"
	//

	for(i = pos; i < pLI->KeysCount; i++)
	{
		dwTempNKoff = HHS + (pLI->Offset)[i];

		(pLI->Offset)[i] = dwNKoff - HHS;

		dwNKoff = dwTempNKoff;
	}

	//
	// Receive name of old key remained without "place"
	//

	pNK = reinterpret_cast<NK_CELL*>(pData + U32_2PTRDIFF(dwNKoff + LFS));

	// To be on the safe side :)
	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::Swap]"
											" Incorrect NK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.81:");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	sKeyName =
		QString::fromAscii(static_cast<char*>(pNK->Name), pNK->NameLength);

	return  PRL_ERR_SUCCESS;
}


/**
 * Updates existing Ri index cell.
 *
 * @author igorl
 *
 * @param[in] szKeyName is the name of new adding key. Now only ASCII names
 * are supported.
 * @param[in] dwNKnewOff is the absolute offset of new adding key.
 * @param[in] dwRiIndexOffsetOld is the relative offset of Ri index cell of
 * parent key.
 * @param[out] dwRiIndexOffsetNew is the absolute offset of updated Ri cell.
 *
 * @return PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
 */
err_status_t CNtBasicFunctions::UpdateRi(
									const char*		szKeyName,
									quint32			dwNKnewOff,
									quint32			dwRiIndexOffsetOld,
									quint32&		dwRiIndexOffsetNew)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateRi]");
	*/

	quint8*			pData = NULL;
	QString			sKeyName = QString(szKeyName);
	// Unicode support
	QByteArray		aKeyNameBuf = sKeyName.toAscii();
	char*			pName = aKeyNameBuf.data();
	quint32			IndexOffset;
	quint32			dwIndexOffsetOld;
	quint32			dwIndexOffsetNew;
	quint32			dwProcessNKoff = dwNKnewOff;
	RI_CELL*		pRI = NULL;
	quint16			ri_pos, IndexType;
	int				cmp, nlen;
	bool			bIsFound, bLast;
	err_status_t	Err = WinRegistry::REG_UNKNOWN_ERROR;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateRi] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.19:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pRI = (RI_CELL*) (pData + U32_2PTRDIFF(HHS + dwRiIndexOffsetOld + LFS));

	if(pRI->Signature != ID_RI)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateRi] Incorrect index cell type.");
		WRITE_TRACE(DBG_FATAL, "OA00002.20:");

		return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
	}

	// NB!
	// Offset of Ri index cell may be not changed. In this case one of
	// subordinate indexes was changed.
	dwRiIndexOffsetNew = dwRiIndexOffsetOld + HHS;

	ri_pos = 0;

Lbl_Repeat: ;

	//
	// 1. Looking for place of the key with name pName and offset dwProcessNKoff.
	//    NB!
	//    - This key may differ from starting key if this starting key hits to full
	//      subordinate Li index cell.
	//    - Value of ri_pos variable is not changed for cutting down searrch.
	//      May be better to use ri_pos + 1?
	//

	nlen     = strlen(pName);
	bIsFound = false;
	bLast    = false;
	do
	{
		IndexOffset = HHS + pRI->Offset[ri_pos] + LFS;
		IndexType   = *((quint16*) (pData + U32_2PTRDIFF(IndexOffset)));

		if(IndexType == ID_LF || IndexType == ID_LH)
		{
			LF_LH_CELL	*pLF = (LF_LH_CELL*) (pData + U32_2PTRDIFF(IndexOffset));
			NK_CELL		*pNK = NULL;

			//
			// Check hit of new key in this "indexes array". These "indexes" are arranged
			// by lexicographical order. Therefore I may make comparision only with first
			// and last keys.
			//

			// First key
			pNK = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + (pLF->Hash[0]).Offset + LFS));

			// FIXME: This comparision is case-insensitive. On Vista I have some
			// problem because of this.
			// May be already fixed (see UCMDB20485)?

			cmp = Compare(pName, pNK->Name, (nlen > pNK->NameLength) ? pNK->NameLength : nlen);

			if(cmp < 0 || (cmp == 0 && nlen < pNK->NameLength))
			{
				bIsFound = true;
				dwIndexOffsetOld = IndexOffset - HHS - LFS;
				break;
			}

			// Last key
			pNK = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + (pLF->Hash[pLF->KeysCount-1]).Offset + LFS));

			// FIXME: This comparision is case-insensitive. On Vista I have some
			// problem because of this.
			// May be already fixed (see UCMDB20485)?

			cmp = Compare(pName, pNK->Name, (nlen > pNK->NameLength) ? pNK->NameLength : nlen);

			if(cmp < 0 || (cmp == 0 && nlen < pNK->NameLength))
			{
				bIsFound = true;
				dwIndexOffsetOld = IndexOffset - HHS - LFS;
				break;
			}
		}

		if(IndexType == ID_LI)
		{
			LI_CELL		*pLI = (LI_CELL*) (pData + U32_2PTRDIFF(IndexOffset));
			NK_CELL		*pNK = NULL;

			//
			// Check hit of new key in this "indexes array". These "indexes" are arranged
			// by lexicographical order. Therefore I may make comparision only with first
			// and last keys.
			//

			// First key
			pNK = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + pLI->Offset[0] + LFS));

			// FIXME: This comparision is case-insensitive. On Vista I have some
			// problem because of this.
			// May be already fixed (see UCMDB20485)?

			cmp = Compare(pName, pNK->Name, (nlen > pNK->NameLength) ? pNK->NameLength : nlen);

			if(cmp < 0 || (cmp == 0 && nlen < pNK->NameLength))
			{
				bIsFound = true;
				dwIndexOffsetOld = IndexOffset - HHS - LFS;
				break;
			}

			// Last key
			pNK = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + pLI->Offset[pLI->KeysCount-1] + LFS));

			// FIXME: This comparision is case-insensitive. On Vista I have some
			// problem because of this.
			// May be already fixed (see UCMDB20485)?

			cmp = Compare(pName, pNK->Name, (nlen > pNK->NameLength) ? pNK->NameLength : nlen);

			if(cmp < 0 || (cmp == 0 && nlen < pNK->NameLength))
			{
				bIsFound = true;
				dwIndexOffsetOld = IndexOffset - HHS - LFS;
				break;
			}
		}

		ri_pos++;
	}
	while(!bIsFound && ri_pos < pRI->IndexesCount);

	if(!bIsFound)
	{
		// New key hits in last "indexes array" and will occupy last position
		// int it

		dwIndexOffsetOld = IndexOffset - HHS - LFS;
		ri_pos           = pRI->IndexesCount - 1;
	}

	if(ri_pos == pRI->IndexesCount - 1)
		// Adding key will be placed in the last index cell
		bLast = true;

	//
	// 2. Update "indexes array"
	//

	if(IndexType == ID_LF || IndexType == ID_LH)
	{
		// We try to add new key to Lf/Lh cell. This is simple case because
		// if this cell is full then we convert it to Li cell simple.

		Err = UpdateLfLh(pName, dwProcessNKoff, dwIndexOffsetOld, dwIndexOffsetNew);

		if(Err == WinRegistry::REG_INDEXES_ARRAY_FULL_ERROR)
		{
			// For statistics collection
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateRi] CONVERTING LF/LH->LI.");
			WRITE_TRACE(DBG_FATAL, "OA00002.21:");

			Err = ConvertLfLhToLi(pName, dwProcessNKoff, dwIndexOffsetOld, dwIndexOffsetNew);
		}

		// From here we "go to" return from this function.
	}
	else
	{
		if(IndexType == ID_LI)
		{
			Err = UpdateLi(pName, dwProcessNKoff, dwIndexOffsetOld, dwIndexOffsetNew);

			if(Err == WinRegistry::REG_INDEXES_ARRAY_FULL_ERROR)
			{
				if(bLast)
				{
					// We try to add new key to last subordinate index cell. But this
					// cell is "Li" type and this cell is full. Thus we need to create
					// new "Lf/Lh" cell and rebuild Ri index cell itself.

					// For statistics collection
					LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateRi] REBUILD RI.");
					WRITE_TRACE(DBG_FATAL, "OA00002.22:");

					Err = CreateLfLh(pName, dwProcessNKoff, dwIndexOffsetNew);

					if(PRL_FAILED(Err))
						return Err;

					// From here we may do "return" because rebuilding was done.

					return  AddLfLhToRi(dwRiIndexOffsetOld, dwIndexOffsetNew, dwRiIndexOffsetNew);
				}
				else
				{
					// We try to add new key to some index cell. But this cell is
					// "Li" type and this cell is full, i.e. when we place new key
					// to this cell we receive some old key for placing in next
					// index cell.

					// For statistics collection
					LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateRi]"
										" Swap %s 0x%x.", qPrintable(sKeyName), dwProcessNKoff);

					// Swapping new key to the old one
					Err = Swap(sKeyName, dwProcessNKoff, dwIndexOffsetOld);

					if(PRL_FAILED(Err))
						return Err;

					// For statistics collection
					LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateRi]"
										" Swap %s 0x%x.", qPrintable(sKeyName), dwProcessNKoff);

					// Update key name and
					aKeyNameBuf = sKeyName.toAscii();
					pName = aKeyNameBuf.data();

					// Place old key to new place
					goto Lbl_Repeat;
				}
			}
		}
		else
		{
			// We MUST not be here

			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::UpdateRi]"
								" Something very very bad (0x%x).", IndexType);
			WRITE_TRACE(DBG_FATAL, "OA00002.82:\t0x%x", IndexType);
		}
	}

	if(PRL_SUCCEEDED(Err))
	{
		// Updating of "indexes array" was successful, save its new offset

		pRI->Offset[ri_pos] = dwIndexOffsetNew - HHS;

		return  PRL_ERR_SUCCESS;
	}

	return  Err;
}


/**
 * Creates Li index cell based on Lf/Lh index cell. Previous Lf/Lh
 * index cell will be freed in case of success creation of new index
 * cell.
 * NB!
 * You need in this function when UpdateLfLh returns
 * WinRegistry::REG_INDEXES_ARRAY_FULL_ERROR error.
 *
 * @author igorl
 *
 * @param[in] szKeyName is the name of new created key.
 * This name is used for determination of correct placement
 * the created key in index cell.
 * @param[in] dwNKnewOff is the offset of new created key.
 * The offset must be absolute.
 * @param[in] dwIndexOffsetOld is the offset of previous
 * index cell. The offset must be relative.
 * @param[out] dwIndexOffsetNew is the offset of new created
 * Li index cell. This offset is absolute.
 *
 * @return PRL_ERR_SUCCESS if there are no error or detailed error
 * code otherwise.
 */
err_status_t CNtBasicFunctions::ConvertLfLhToLi(
									const char*		szKeyName,
									quint32			dwNKnewOff,
									quint32			dwIndexOffsetOld,
									quint32&		dwIndexOffsetNew)
{
	// Code of this function is base on UpdateLi function.

	/*
	LOG_MESSAGE(DBG_FATAL, "[ConvertLfLhToLi]");
	*/

	quint8*			pData = NULL;
	quint32			size;
	LF_LH_CELL*		pLFold = NULL;
	LI_CELL*		pLI = NULL;
	int				nlen, cmp, pos;
	quint16			i, j;
	err_status_t	err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::ConvertLfLhToLi]"
														" Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.70:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pLFold = reinterpret_cast<LF_LH_CELL*>(pData + U32_2PTRDIFF(HHS + dwIndexOffsetOld + LFS));

	if(pLFold->Signature != ID_LF && pLFold->Signature != ID_LH)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::ConvertLfLhToLi]"
												" Incorrect index cell type.");
		WRITE_TRACE(DBG_FATAL, "OA00002.71:");

		return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
	}

	nlen  = strlen(szKeyName);

	// Size of new "indexes array" (it contains additional new index)
	// P.S.
	// sizeof(quint32) - size of offset filed in LI_CELL struture
	size = sizeof(LI_CELL) + sizeof(quint32) * pLFold->KeysCount;

	// Allocate space for new "indexes array"
	err = m_pRawData->AllocCell(size, dwIndexOffsetNew);

	if(PRL_FAILED(err))
		return err;

	pLI = reinterpret_cast<LI_CELL*>(pData + U32_2PTRDIFF(dwIndexOffsetNew + LFS));

	// Index cell type
	pLI->Signature = ID_LI;
	// Count of addressed keys
	pLI->KeysCount = pLFold->KeysCount + 1;

	//
	// The following loop is meaned for
	// 1. Looking for position of new key in "indexes array". These "indexes"
	//    are arranged by lexicographical order.
	// 2. Copying old "indexes array" to new one. Element for new key will be
	//    empty.
	//

	for(i = 0, j = 0, pos = -1; i < pLFold->KeysCount; i++)
	{
		// Looking for position of new key
		if(pos == -1)
		{
			NK_CELL*  pNK = reinterpret_cast<NK_CELL*>
								(pData + U32_2PTRDIFF(HHS + ((pLFold->Hash)[i]).Offset + LFS));

			// FIXME: This comparision is case-insensitive. On Vista I have some
			// problem because of this.
			// May be already fixed (see UCMDB20485)?

			cmp = Compare(szKeyName, pNK->Name,
								(nlen > pNK->NameLength) ? pNK->NameLength : nlen);

			if(cmp == 0 && nlen == pNK->NameLength)
			{
				LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::ConvertLfLhToLi]"
											" Key '%s' already exists.", szKeyName);
				WRITE_TRACE(DBG_FATAL, "OA00002.72:\t%s", szKeyName);

				m_pRawData->FreeCell(dwIndexOffsetNew);

				return  WinRegistry::REG_KEY_ALREADY_EXISTS_ERROR;
			}

			if(cmp < 0 || (cmp == 0 && nlen < pNK->NameLength))
			{
				// Position was found

				pos = i;
				// Skip position of new key
				j++;
			}
		}

		// Copying element of old "indexes array"
		(pLI->Offset)[j] = ((pLFold->Hash)[i]).Offset;

		j++;
	}

	if(pos < 0)
		pos = pLFold->KeysCount;

	// Offset of key itself
	(pLI->Offset)[pos] = dwNKnewOff - HHS;

	// Free old "indexes array"
	return  m_pRawData->FreeCell(dwIndexOffsetOld + HHS);
}


/**
 * Creates Ri index cell based on Li index cell. Previous Li
 * index cell will be re-used.
 * NB!
 * You need in this function when UpdateLi returns
 * WinRegistry::REG_INDEXES_ARRAY_FULL_ERROR error.
 *
 * @author igorl
 *
 * @param[in] szKeyName is the name of new created key.
 * This name is used for determination of correct placement
 * the created key in index cell.
 * @param[in] dwNKnewOff is the offset of new created key.
 * The offset must be absolute.
 * @param[in] dwLiOff is the offset of previous index cell
 * The offset must be relative.
 * @param[out] dwRiOff is the offset of new created Ri index
 * cell. This offset is absolute.
 *
 * @return PRL_ERR_SUCCESS if there are no error or detailed error
 * code otherwise.
 */
err_status_t CNtBasicFunctions::ConvertLiToRi(
									const char*		szKeyName,
									quint32			dwNKnewOff,
									quint32			dwLiOff,
									quint32&		dwRiOff)
{
	// Idea is very simple. We made Ri index cell from existed Li
	// cell. After that we call UpdateRi for new created key.

	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::ConvertLiToRi]");
	*/

	quint8*			pData  = NULL;
	quint32			size;
	RI_CELL*		pRI    = NULL;
	LI_CELL*		pLI    = NULL;
	err_status_t	err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::ConvertLiToRi]"
													" Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.83:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pLI = reinterpret_cast<LI_CELL*>(pData + U32_2PTRDIFF(HHS + dwLiOff + LFS));

	if(pLI->Signature != ID_LI)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::ConvertLiToRi]"
											" Index cell type is not LI.");
		WRITE_TRACE(DBG_FATAL, "OA00002.84:");

		return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
	}

	if(pLI->KeysCount != MAX_LI_COUNT)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::ConvertLiToRi]"
											" 'Indexes array' is not full.");
		WRITE_TRACE(DBG_FATAL, "OA00002.85:");

		// Not very good, but ...
		return  WinRegistry::REG_UNKNOWN_ERROR;
	}

	// Size of new "indexes array" (it contains additional new index)
	// P.S.
	// sizeof(quint32) - size of offset filed in RI_CELL struture
	size = sizeof(RI_CELL);

	// Allocate space for new "indexes array"
	err  = m_pRawData->AllocCell(size, dwRiOff);

	if(PRL_FAILED(err))
		return err;

	pRI = reinterpret_cast<RI_CELL*>(pData + U32_2PTRDIFF(dwRiOff + LFS));

	// Index cell type
	pRI->Signature = ID_RI;
	// Count of addressed keys
	pRI->IndexesCount = 1;
	// Offsets
	(pRI->Offset)[0] = dwLiOff;

	return  UpdateRi(szKeyName, dwNKnewOff, dwRiOff - HHS, dwRiOff);
}


//
// Function create/update existing "indexes array"
//
// Parameters:
//    szKeyName      [in]  - name of key. We will calculate hash value on base of this name. NB: now we examine
//                           only ASCII names.
//    NKnewOff       [in]  - absolute offset of cell (!) in which this key was placed key.
//    IndexOffsetOld [in]  - relative offset of old "indexes array".  This array will be cleared out in case of
//                           its successful updating but value of this offset will not be changed (!).
//    IndexOffsetNew [out] - absolute offset of new cell (!) in which updated "indexes array" will be placed. If
//                           some error happens it will be equal to -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::RebuildIndex(const char* szKeyName, quint32 NKnewOff, quint32 IndexOffsetOld, quint32& IndexOffsetNew)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::RebuildIndex]");
	*/

	if(IndexOffsetOld == 0 || IndexOffsetOld == (quint32) -1)
	{
		// Parent key does not contain any child, create new "indexes array"

		return  CreateLfLh(szKeyName, NKnewOff, IndexOffsetNew);
	}

	quint8	*pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::RebuildIndex] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.23:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	quint16	IndexType = *((quint16*) (pData + U32_2PTRDIFF(HHS + IndexOffsetOld + LFS)));

	switch(IndexType)
	{
		case ID_LF:
		case ID_LH: {
						err_status_t  Err = UpdateLfLh(szKeyName, NKnewOff, IndexOffsetOld, IndexOffsetNew);

						if(Err == WinRegistry::REG_INDEXES_ARRAY_FULL_ERROR)
						{
							// For statistics collection
							LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::RebuildIndex] CONVERTING LF/LH->LI.");
							WRITE_TRACE(DBG_FATAL, "OA00002.24:");

							Err = ConvertLfLhToLi(szKeyName, NKnewOff, IndexOffsetOld, IndexOffsetNew);
						}

						return  Err;
					}

					break;

		case ID_LI: {
						err_status_t  Err = UpdateLi(szKeyName, NKnewOff, IndexOffsetOld, IndexOffsetNew);

						if(Err == WinRegistry::REG_INDEXES_ARRAY_FULL_ERROR)
						{
							// For statistics collection
							LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::RebuildIndex] CONVERTING LI->RI.");
							WRITE_TRACE(DBG_FATAL, "OA00002.25:");

							Err = ConvertLiToRi(szKeyName, NKnewOff, IndexOffsetOld, IndexOffsetNew);
						}

						return  Err;
					}

					break;

		case ID_RI: return  UpdateRi(szKeyName, NKnewOff, IndexOffsetOld, IndexOffsetNew);
	};

	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::RebuildIndex] Incorrect index cell type.");
	WRITE_TRACE(DBG_FATAL, "OA00002.26:");

	return  WinRegistry::REG_INCORRECT_INDEX_CELL_ERROR;
}

//
// Function creates new key.
//
// Parameters:
//    sKeyName    [in]  - name of key. NB: now we examine only ASCII names.
//    dwKeyOffset [out] - absolute offset of created key. If some error happens it will be equal to -1.
//    dwOffset    [in]  - absolute offset of parent key. Exactly in this key we will create subkey with given
//                        name. If we want to create subkey in root key we follow to set this offset equal -1.
//                        This parameter is not obligatory, its default value is equal to -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::CreateKey(const QString& sKeyName, quint32& dwKeyOffset, quint32 dwOffset)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateKey]");
	*/

	quint8			*pData = NULL;
	NK_CELL			*pNK = NULL;
	NK_CELL			*pNKnew = NULL;
	quint32			NKnewOff;
	quint32			dwOffIndexNew;
	err_status_t	Err;

	dwKeyOffset = (quint32) -1;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::RebuildIndex] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.27:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	if(sKeyName.isEmpty())
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateKey] Key name is empty.");
		WRITE_TRACE(DBG_FATAL, "OA00002.28:");

		return  WinRegistry::REG_KEY_NAME_EMPTY_ERROR;
	}

	if(dwOffset == (quint32) -1)
		// Offset was not determine, lets go from root key
		dwOffset = HHS + m_pRawData->GetRootCellOffset() + LFS;

	// pNK - in this key we shall add sub key
	pNK = (NK_CELL*) (pData + U32_2PTRDIFF(dwOffset));

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateKey] Incorrect NK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.29:");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	// Allocate space for new key.
	// P.S.
	// -1 - because one symbol already define in structure NK_CELL
	Err = m_pRawData->AllocCell(sizeof(NK_CELL) + sKeyName.size() - 1, NKnewOff);

	if(PRL_FAILED(Err))
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateKey] Can not allocate space (0x%x bytes) for 'indexes array' (error code: %d).", sizeof(NK_CELL) + sKeyName.size() - 1, Err);
		WRITE_TRACE(DBG_FATAL, "OA00002.30:\t%x;\t%d", sizeof(NK_CELL) + sKeyName.size() - 1, Err);

		return  Err;
	}

	pNKnew = (NK_CELL*) (pData + U32_2PTRDIFF(NKnewOff + LFS));

	//
	// Fill NK cell of new subkey
	//

	pNKnew->Signature		= ID_NK;
	// 0x20 - key name is not in UNICODE, FIXME: Need to add constant
	pNKnew-> Type			= 0x20;
	pNKnew->ParentOffset	= dwOffset - HHS - LFS;
	pNKnew->SubKeysCount	= 0;
	pNKnew->OffSubKeysIndex	= (quint32) -1;
	pNKnew->Other3			= static_cast<quint32>(-1);
	pNKnew->ValuesCount		= 0;
	pNKnew->OffValuesIndex	= (quint32) -1;
	pNKnew->OffSk			= pNK->OffSk;
	pNKnew->OffClassName	= (quint32) -1;
	pNKnew->NameLength		= sKeyName.size();
	pNKnew->ClassNameLength	= 0;
	memcpy(pNKnew->Name, sKeyName.toAscii().data(), sKeyName.size());

	// Some notes according to pNKnew:
	// - I (igorl) fill field "TimeStap" by zeros;
	// - I use parent security key in created sub-key.

	Err = RebuildIndex(sKeyName.toAscii().data(), NKnewOff, pNK->OffSubKeysIndex, dwOffIndexNew);

	if(PRL_FAILED(Err))
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateKey] Can not rebuild 'indexes array' (error code: %d).", Err);
		WRITE_TRACE(DBG_FATAL, "OA00002.31:\t%d", Err);

		m_pRawData->FreeCell(NKnewOff);

		return  Err;
	}

	//
	// Update some fields in parent key
	//

	// Count of sub-keys
	pNK->SubKeysCount += 1;

	// Offset of updated index cell
	pNK->OffSubKeysIndex = dwOffIndexNew - HHS;

	// "Usage count" of corresponded SK cell
	SK_CELL*  pSK = reinterpret_cast<SK_CELL*>(pData + U32_2PTRDIFF(HHS + pNK->OffSk + LFS));

	if(pSK->Signature == ID_SK)
		pSK->UsageCount += 1;
	else
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateKey]"
						" Parent key has incorrect SK (0x%x).", pSK->Signature);
		WRITE_TRACE(DBG_FATAL, "OA00002.86:\t0x%x", pSK->Signature);
	}

	//
	// This comment is used for information purpose. See comment to NL_CELL::Other4 also.
	//
	// At this moment we have strange effect of key creation. If you create "TestUp\TestDown"
	// anywhere, load changed hive to your registry by help of "reg" and try to delete "TestUp"
	// by help of "regedit" then you can not do it on unknown reason(*). But you can delete
	// "TestUp" by help of Windows API (see example "Deleting a Key with Subkeys" from MSDN).
	// Besides that the following "work around" exists. You may create one more sub-key in
	// "TestUp" by help of "regedit" and after that "regedit" can delete "TestUp".
	// (*)
	// Using command "reg DELETE" gives the following results:
	// X:\...>reg DELETE HKLM\TempHive\Research
	// Permanently delete the registry key TempHive\Research (Y/N)? Y
	// Error:  More data is available.
	//

	quint32  dwLen = static_cast<quint32>(sKeyName.size()) * 2;

	pNK->Other4 = (pNK->Other4 <  dwLen) ? dwLen : pNK->Other4;

	// We do nothing with pNK->Other5 because we do not use "Classes"

	dwKeyOffset = NKnewOff + LFS;

	return  PRL_ERR_SUCCESS;
}

//
// Function deletes subkey.
//
// Parameters:
//    sKeyName [in] - name of key. NB: now we examine only ASCII names.
//    dwOffset [in] - absolute offset of parent key. Exactly in this key we will create subkey with given
//                    name.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
// P.S.
// This function is very large. I may break it to several small functions, but these functions
// will be used only this function and their purpose will be strained to my mind.
//
err_status_t CNtBasicFunctions::DeleteKey(const QString& sKeyName, quint32 dwOffset)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteKey]");
	*/

	quint8			*pData = NULL;
	NK_CELL			*pNK = NULL;
	NK_CELL			*pNKdel = NULL;
	quint32			NKdelOff, IndexOffset;
	quint16			pos, ri_pos, IndexType;
	LF_LH_CELL		*pLF = NULL;
	quint32			LFoff = 0;
	LF_LH_CELL		*pLFnew = NULL;
	quint32			LFnewOff = 0;
	quint32			LFsize = 0;
	LI_CELL			*pLI = NULL;
	quint32			LIoff = 0;
	LI_CELL			*pLInew = NULL;
	quint32			LInewOff = 0;
	quint32			LIsize = 0;
	RI_CELL			*pRI = NULL;
	quint32			RIoff = 0;
	RI_CELL			*pRInew = NULL;
	quint32			RInewOff = 0;
	quint32			RIsize = 0;
	SK_CELL			*pSK = NULL;
	err_status_t	Err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteKey] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.32:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	if(dwOffset == quint32(-1))
		// Offset was not determine, lets go from root key
		dwOffset = HHS + m_pRawData->GetRootCellOffset() + LFS;

	pNK = (NK_CELL*) (pData + U32_2PTRDIFF(dwOffset));

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteKey] Incorrect NK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.33:");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	if(!pNK->SubKeysCount)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteKey] Key has no subkeys.");
		WRITE_TRACE(DBG_FATAL, "OA00002.34:");

		return  WinRegistry::REG_KEY_NO_SUBKEY_ERROR;
	}

	//
	// 1. Look for position of deleting key
	//

	NKdelOff = FindSubKey(pNK, sKeyName.toAscii().data(), pos, ri_pos);

	if(NKdelOff == (quint32) -1)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteKey] Subkey was not found.");
		WRITE_TRACE(DBG_FATAL, "OA00002.35:");

		return  WinRegistry::REG_KEY_NOT_FOUND_ERROR;
	}

	// This key will be deleted
	pNKdel = (NK_CELL*) (pData + U32_2PTRDIFF(NKdelOff));

	if(pNKdel->ValuesCount || pNKdel->SubKeysCount)
	{
		LOG_MESSAGE(DBG_FATAL, "[[CNtBasicFunctions::DelKey]] Subkey is not empty.");
		WRITE_TRACE(DBG_FATAL, "OA00002.36:");

		// It must be empty
		return  WinRegistry::REG_KEY_NOT_EMPTY_ERROR;
	}

	//
	// 2. Now we should delete record about deleting key from "indexes array"
	//

	IndexType = *((quint16*) (pData + U32_2PTRDIFF(HHS + pNK->OffSubKeysIndex + LFS)));

	if(IndexType == ID_RI)
	{
		pRI   = (RI_CELL*) (pData + U32_2PTRDIFF(HHS + pNK->OffSubKeysIndex + LFS));
		RIoff = HHS + pNK->OffSubKeysIndex + LFS;
	}
	else
		pRI = NULL;

	if(pRI)
	{
		IndexOffset = HHS + pRI->Offset[ri_pos] + LFS;
		IndexType   = *((quint16*) (pData + U32_2PTRDIFF(IndexOffset)));
	}
	else
		IndexOffset = HHS + pNK->OffSubKeysIndex + LFS;

	if(IndexType == ID_LF || IndexType == ID_LH)
	{
		pLF   = (LF_LH_CELL*) (pData + U32_2PTRDIFF(IndexOffset));
		LFoff = IndexOffset;

		if(pLF->KeysCount == 1)
		{
			// We will delete only one key

			pLFnew = NULL;
			LFsize = 0;
		}
		else
		{
			// FIXME: size of allocated memory was greater than value of LFsize
			LFsize = sizeof(LF_LH_CELL) + sizeof(LF_LH_CELL::LF_LH_HASH) * (pLF->KeysCount - 2);
			pLFnew = (LF_LH_CELL*) new quint8 [LFsize];

			//
			// Form new "indexes array"
			//

			pLFnew->Signature = pLF->Signature;
			pLFnew->KeysCount = pLF->KeysCount - 1;

			for(quint16 i = 0, j = 0; i < pLF->KeysCount; i++)
			{
				if(i != pos)
				{
					(pLFnew->Hash)[j].Offset = (pLF->Hash)[i].Offset;
					memcpy((pLFnew->Hash)[j].Name, (pLF->Hash)[i].Name, sizeof((pLF->Hash)[i].Name));

					j++;
				}
			}
		}
	}

	if(IndexType == ID_LI)
	{
		pLI   = (LI_CELL*) (pData + U32_2PTRDIFF(IndexOffset));
		LIoff = IndexOffset;

		if(pLI->KeysCount == 1)
		{
			// We will delete only one key

			pLInew = NULL;
			LIsize = 0;
		}
		else
		{
			// FIXME: size of allocated memory was greater than value of LIsize
			LIsize = sizeof(LI_CELL) + sizeof(quint32) * (pLI->KeysCount - 2);
			pLInew = (LI_CELL*) new quint8 [LIsize];

			//
			// Form new "indexes array"
			//

			pLInew->Signature = pLI->Signature;
			pLInew->KeysCount = pLI->KeysCount - 1;

			for(quint16 i = 0, j = 0; i < pLI->KeysCount; i++)
			{
				if(i != pos)
				{
					(pLInew->Offset)[j] = (pLI->Offset)[i];

					j++;
				}
			}
		}
	}

	if(pRI)
	{
		if(pRI->IndexesCount == 1 && !pLFnew && !pLInew)
		{
			// We will delete only one key

			pRInew = NULL;
			RIsize = 0;
		}
		else
		{
			if(!pLFnew && !pLInew)
			{
				//
				// After deleting key "indexes array" is empty and we should exclude it
				// from RI "indexes array"
				//

				// FIXME: size of allocated memory was greater than value of RIsize
				RIsize = sizeof(RI_CELL) + sizeof(quint32) * (pRI->IndexesCount - 2);
				pRInew = (RI_CELL*) new quint8 [RIsize];

				pRInew->Signature    = pRI->Signature;
				pRInew->IndexesCount = pRI->IndexesCount - 1;

				//
				// Form new "indexes array"
				//

				for(quint16 i = 0, j = 0; i < pRI->IndexesCount; i++)
				{
					if(i != ri_pos)
					{
						(pRInew->Offset)[j] = (pRI->Offset)[i];

						j++;
					}
				}
			}
		}
	}

	//
	// 3. Allocate memory for new "indexes array"
	//

	if(pRInew)
	{
		Err = m_pRawData->AllocCell(RIsize, RInewOff);

		if(PRL_FAILED(Err))
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteKey] Can not allocate space (0x%x bytes) for 'indexes array' (error code: %d).", RIsize, Err);
			WRITE_TRACE(DBG_FATAL, "OA00002.37:\t%x;\t%d", RIsize, Err);

			if(pRInew)
				delete[] pRInew;
			if(pLInew)
				delete[] pLInew;
			if(pLFnew)
				delete[] pLFnew;

			return  Err;
		}
	}

	// NB!
	// One of pointers pLInew or pLFnew is NULL necessarily

	if(pLInew)
	{
		Err = m_pRawData->AllocCell(LIsize, LInewOff);

		if(PRL_FAILED(Err))
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DelKey] Can not allocate space (0x%x bytes) for 'indexes array' (error code: %d).", LIsize, Err);
			WRITE_TRACE(DBG_FATAL, "OA00002.38:\t%x;\t%d", LIsize, Err);

			if(pRInew)
				m_pRawData->FreeCell(RInewOff);

			if(pRInew)
				delete[] pRInew;
			if(pLInew)
				delete[] pLInew;
			if(pLFnew)
				delete[] pLFnew;

			return  Err;
		}
	}

	if(pLFnew)
	{
		Err = m_pRawData->AllocCell(LFsize, LFnewOff);

		if(PRL_FAILED(Err))
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DelKey] Can not allocate space (0x%x bytes) for 'indexes array' (error code: %d).", LFsize, Err);
			WRITE_TRACE(DBG_FATAL, "OA00002.39:\t%x;\t%d", LFsize, Err);

			if(pRInew)
				m_pRawData->FreeCell(RInewOff);

			if(pRInew)
				delete[] pRInew;
			if(pLInew)
				delete[] pLInew;
			if(pLFnew)
				delete[] pLFnew;

			return  Err;
		}
	}

	//
	// 4. Copy new data from memory to registry hive
	//

	// FIXME Replace magic-numbers with constan
	if(pRInew)
		memcpy((RI_CELL*) (pData + U32_2PTRDIFF(RInewOff + 4)), pRInew, RIsize);

	if(pLInew)
		memcpy((LI_CELL*) (pData + U32_2PTRDIFF(LInewOff + 4)), pLInew, LIsize);

	if(pLFnew)
		memcpy((LF_LH_CELL*) (pData + U32_2PTRDIFF(LFnewOff + 4)), pLFnew, LFsize);

	//
	// 5. Correct counter and sub-key offset in paren key
	//

	pNK->SubKeysCount = pNK->SubKeysCount - 1;

	if(pRI)
	{
		/*
		if(pRInew == NULL && pRI->IndexesCount == 1)
		*/
		// corrected by mityak@
		if(pRInew == NULL && pRI->IndexesCount == 1 && !pLInew && !pLFnew)
			pNK->OffSubKeysIndex = -1;
		else
		{
			if(!pLInew && !pLFnew)
			{
				pNK->OffSubKeysIndex = RInewOff - HHS;
			}
			else
			{
				if(LInewOff)
					pRI->Offset[ri_pos] = LInewOff - HHS;
				if(LFnewOff)
					pRI->Offset[ri_pos] = LFnewOff - HHS;
			}
		}
	}
	else
	{
		if(LInewOff)
			pNK->OffSubKeysIndex = LInewOff - HHS;
		if(LFnewOff)
			pNK->OffSubKeysIndex = LFnewOff - HHS;

		if(!pLInew && !pLFnew && pNK->SubKeysCount == 0)
			pNK->OffSubKeysIndex = -1;
	}

	//
	// 6. Free all useless areas and memoryes
	//

	if(RIoff && pRInew)
		m_pRawData->FreeCell(RIoff - LFS);

	if(LIoff && pLInew)
		m_pRawData->FreeCell(LIoff - LFS);

	if(LFoff && pLFnew)
		m_pRawData->FreeCell(LFoff - LFS);

	if(pNKdel->ClassNameLength)
		m_pRawData->FreeCell(pNKdel->OffClassName + HHS);

	// Decrease counter of secutity key. If counter is zero then we should delete
	// this security key
	pSK	= (SK_CELL*) (pData + U32_2PTRDIFF(HHS + pNKdel->OffSk + LFS));

	// To be on the safe side
	if(pSK->Signature == ID_SK)
	{
		pSK->UsageCount--;

		if(!pSK->UsageCount)
		{
			SK_CELL		*pSKswap;
			quint32		SKoff;

			SKoff = pSK->OffPrevSk;

			pSKswap = (SK_CELL*) (pData + U32_2PTRDIFF(HHS + SKoff + LFS));

			if(pSKswap->Signature == ID_SK)
				pSKswap->OffNextSk = pSK->OffNextSk;

			SKoff = pSK->OffNextSk;

			pSKswap = (SK_CELL*) (pData + U32_2PTRDIFF(HHS + SKoff + LFS));

			if(pSKswap->Signature == ID_SK)
				pSKswap->OffPrevSk = pSK->OffPrevSk;

			m_pRawData->FreeCell(HHS + pNKdel->OffSk);
		}
	}

	// At last delete key itself
	m_pRawData->FreeCell(NKdelOff - LFS);

	if(pRInew)
		delete[] pRInew;
	if(pLInew)
		delete[] pLInew;
	if(pLFnew)
		delete[] pLFnew;

	/*
	m_bDirty = true;
	*/

	return  PRL_ERR_SUCCESS;
}

//
// Function retrieves information about a specified registry key.
//
// Parameters:
//    dwOffset         [in]  - absolute offset of specified key
//    psClassName      [out] - name of class of specified key (optional; not tested, see comments below)
//    pcSubKeys        [out] - number of subkeys contained by the specified key (optional)
//    pMaxSubKeyLen    [out] - pointer to a variable that receives the length, in characters, of the key's
//                             subkey with the longest name (optional)
//    pMaxClassLen     [out] - pointer to a variable that receives the length, in characters, of the longest
//                             string specifying a subkey class (optional)
//    pcValues         [out] - pointer to a variable that receives the number of values associated with the
//                             key (optional)
//    pMaxValueLen     [out] - pointer to a variable that receives the length, in characters, of the key's
//                             longest value name (optional)
//    pMaxValueDataLen [out] - pointer to a variable that receives the length, in bytes, of the longest data
//                             component among the values of the key (optional)
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::QueryInfoKey(quint32 dwOffset, QString* psClassName, quint32* pcSubKeys, quint16* pMaxSubKeyLen, quint16* pMaxClassLen, quint32* pcValues, quint16* pMaxValueLen, quint32* pMaxValueDataLen)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::QueryInfoKey]");
	*/

	quint8		*pData = NULL;
	NK_CELL		*pNK = NULL;
	quint16		IndexType;
	quint32		IndexOffset;
	quint16		pos;
	quint16		ri_pos;
	RI_CELL		*pRI = NULL;
	VK_CELL		*pVK = NULL;
	quint32		*pValuesOffset = NULL;
	quint32		i;
	QString		sClassName;
	quint32		cSubKeys;
	quint16		MaxSubKeyLen = 0;
	quint16		MaxClassLen  = 0;
	quint32		cValues = 0;
	quint16		MaxValueLen = 0;
	quint32		MaxValueDataLen = 0;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::QueryInfoKey] Storage is NULL.");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	if(dwOffset == quint32(-1))
		// Offset was not determine, lets go from root key
		dwOffset = HHS + m_pRawData->GetRootCellOffset() + LFS;

	pNK = (NK_CELL*) (pData + dwOffset);

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::QueryInfoKey] Incorrect NK cell.");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	//
	// 1. Some information from specified key.
	//

	if(pNK->OffClassName != 0 && pNK->OffClassName != (quint32) -1)
	{
		//
		// NB!
		// I do not test the following because I can not find key belonged to some class.
		//

		const char	*pBuf = (const char*) (pData + U32_2PTRDIFF(HHS + pNK->OffClassName + LFS));

		sClassName.setUtf16((const ushort *) pBuf, pNK->ClassNameLength);
	}
	else
		sClassName = QString();

	cSubKeys = pNK->SubKeysCount;
	cValues  = pNK->ValuesCount;

	//
	// 2. Loop on subkeys (determing of MaxSubKeyLen and MaxClassLen).
	//
	// P.S.
	// The same loop already exists in FindSubKey.
	//

	// Determine child index type
	IndexType = *((quint16*) (pData + U32_2PTRDIFF(HHS + pNK->OffSubKeysIndex + LFS)));

	if(IndexType == ID_RI)
	{
		pRI    = (RI_CELL*) (pData + U32_2PTRDIFF(HHS + pNK->OffSubKeysIndex + LFS));
		ri_pos = 0;
	}
	else
	{
		pRI    = NULL;
		ri_pos = (quint16) -1;
	}

	MaxSubKeyLen = 0;
	MaxClassLen  = 0;
	do
	{
		// Get offset child index
		if(pRI)
		{
			// In case of 'ri' index get next sub child index
			IndexOffset = HHS + pRI->Offset[ri_pos] + LFS;
			IndexType   = *((quint16*) (pData + U32_2PTRDIFF(IndexOffset)));
		}
		else
		{
			IndexOffset = HHS + pNK->OffSubKeysIndex + LFS;
		}

		if(IndexType == ID_LF || IndexType == ID_LH)
		{
			LF_LH_CELL	*pLF = (LF_LH_CELL*) (pData + U32_2PTRDIFF(IndexOffset));

			// Runing on subkeys
			pos = 0;
			while(pos < pLF->KeysCount)
			{
				NK_CELL		*pNKnext = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + (pLF->Hash[pos]).Offset + LFS));

				// To be on the safe side :)
				if(pNKnext->Signature != ID_NK)
				{
					LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::QueryInfoKey] Incorrect NK cell (lf/lh loop).");

					return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
				}

				if(pNKnext->NameLength > MaxSubKeyLen)
					MaxSubKeyLen = pNKnext->NameLength;
				if(pNKnext->ClassNameLength > MaxClassLen)
					MaxClassLen = pNKnext->ClassNameLength;

				// To next subkey
				pos++;
			}
		}

		if(IndexType == ID_LI)
		{
			LI_CELL		*pLI = (LI_CELL*) (pData + U32_2PTRDIFF(IndexOffset));

			// Runing on subkeys
			pos = 0;
			while(pos < pLI->KeysCount)
			{
				NK_CELL		*pNKnext = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + pLI->Offset[pos] + LFS));

				// To be on the safe side :)
				if(pNKnext->Signature != ID_NK)
				{
					LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::QueryInfoKey] Incorrect NK cell (li loop).");

					return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
				}

				if(pNKnext->NameLength > MaxSubKeyLen)
					MaxSubKeyLen = pNKnext->NameLength;
				if(pNKnext->ClassNameLength > MaxClassLen)
					MaxClassLen = pNKnext->ClassNameLength;

				// To next key
				pos++;
			}
		}

		// To next sub child index
		if(pRI)
			ri_pos++;
	}
	while(pRI && ri_pos < pRI->IndexesCount);	// Loop on array of "indexes arrrayes" in case of RI

	//
	// 3. Loop on values (determing of MaxValueLen and MaxValueDataLen).
	//
	// P.S.
	// The same loop already exists in FindValue.
	//

	pValuesOffset = (quint32*) (pData + U32_2PTRDIFF(HHS + pNK->OffValuesIndex + LFS));

	MaxValueLen     = 0;
	MaxValueDataLen = 0;
	for(i = 0; i < pNK->ValuesCount; i++)
	{
		pVK = (VK_CELL*) (pData + U32_2PTRDIFF(HHS + pValuesOffset[i] + LFS));

		// To be on the safe side :)
		if(pVK->Signature != ID_VK)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::QueryInfoKey] Incorrect VK cell (loop).");

			return  WinRegistry::REG_INCORRECT_VK_CELL_ERROR;
		}

		if(pVK->NameLength > MaxValueLen)
			MaxValueLen = pVK->NameLength;

		//
		// Receive real size of data
		//

		quint32  dwRealDataSize = pVK->DataLength;

		// Special inline case (len = 0x80000000): data (4 bytes?) in type field
		// FIXME: unfortunately but I (igorl) do not remember what is it
		if(pVK->DataLength == 0x80000000)
			dwRealDataSize = sizeof(pVK->Type);
		else
		{
			// Negative len is inline, return ptr to offset-field which in this
			// case contains the data itself
			// FIXME: this may be incorrect, need to test for REG_BINARY and
			// DataLength < 4
			if(pVK->DataLength & 0x80000000)
				dwRealDataSize = sizeof(pVK->OffData);
		}

		if(dwRealDataSize > MaxValueDataLen)
			MaxValueDataLen = dwRealDataSize;
	}

	//
	// 4. Prepare returning values
	//

	if(psClassName)
		*psClassName = sClassName;
	if(pcSubKeys)
		*pcSubKeys = cSubKeys;
	if(pMaxSubKeyLen)
		*pMaxSubKeyLen = MaxSubKeyLen;
	if(pMaxClassLen)
		*pMaxClassLen = MaxClassLen;
	if(pcValues)
		*pcValues = cValues;
	if(pMaxValueLen)
		*pMaxValueLen = MaxValueLen;
	if(pMaxValueDataLen)
		*pMaxValueDataLen = MaxValueDataLen;

	return  PRL_ERR_SUCCESS;
}

//
// Function enumerates subkeys of the specified open registry key.
//
// Parameters:
//    dwOffset [in]  - absolute offset of specified key
//    dwIndex  [in]  - index of the subkey to retrieve
//    psName   [out] - pointer to a QString that receives the name of the subkey (optional)
//    psClass  [out] - pointer to a QString that contains the class of the enumerated subkey
//                     when the function returns (optional; not tested, see comments below)
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::EnumKey(quint32 dwOffset, quint32 dwIndex, QString* psName, QString* psClass)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumKey]");
	*/

	quint8		*pData = NULL;
	NK_CELL		*pNK = NULL;
	quint16		IndexType;
	quint32		IndexOffset;
	quint16		pos;
	quint16		ri_pos;
	RI_CELL		*pRI = NULL;
	quint32		i;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumKey] Storage is NULL.");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	if(dwOffset == quint32(-1))
		// Offset was not determine, lets go from root key
		dwOffset = HHS + m_pRawData->GetRootCellOffset() + LFS;

	pNK = (NK_CELL*) (pData + U32_2PTRDIFF(dwOffset));

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumKey] Incorrect NK cell.");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	//
	// Loop on subkeys (looking for index of the subkey to retrieve).
	//
	// P.S.
	// The same loop already exists in FindSubKey.
	//

	// Determine child index type
	IndexType = *((quint16*) (pData + U32_2PTRDIFF(HHS + pNK->OffSubKeysIndex + LFS)));

	if(IndexType == ID_RI)
	{
		pRI    = (RI_CELL*) (pData + U32_2PTRDIFF(HHS + pNK->OffSubKeysIndex + LFS));
		ri_pos = 0;
	}
	else
	{
		pRI    = NULL;
		ri_pos = (quint16) -1;
	}

	// 'Global' index of current subkey
	i  = 0;
	do
	{
		// Get offset child index
		if(pRI)
		{
			// In case of 'ri' index get next sub child index
			IndexOffset = HHS + pRI->Offset[ri_pos] + LFS;
			IndexType   = *((quint16*) (pData + U32_2PTRDIFF(IndexOffset)));
		}
		else
		{
			IndexOffset = HHS + pNK->OffSubKeysIndex + LFS;
		}

		if(IndexType == ID_LF || IndexType == ID_LH)
		{
			LF_LH_CELL	*pLF = (LF_LH_CELL*) (pData + U32_2PTRDIFF(IndexOffset));

			// Runing on subkeys
			pos = 0;
			while(pos < pLF->KeysCount)
			{
				NK_CELL		*pNKnext = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + (pLF->Hash[pos]).Offset + LFS));

				// To be on the safe side :)
				if(pNKnext->Signature != ID_NK)
				{
					LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumKey] Incorrect NK cell (lf/lh loop).");

					return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
				}

				if(i == dwIndex)
				{
					if(psName)
						*psName = QString(QByteArray(pNKnext->Name, pNKnext->NameLength));

					if(psClass)
					{
						if(pNKnext->OffClassName != 0 && pNKnext->OffClassName != (quint32) -1)
						{
							//
							// NB!
							// I do not test the following because I can not find key belonged to some class.
							//

							const char *pBuf = (const char*)(pData + U32_2PTRDIFF(HHS + pNKnext->OffClassName + LFS));

							*psClass = QString(QByteArray(pBuf, pNKnext->ClassNameLength));
						}
						else
							*psClass = QString();
					}

					return  PRL_ERR_SUCCESS;
				}

				// To next subkey
				pos++;
				// To next 'global' index
				i++;
			}
		}

		if(IndexType == ID_LI)
		{
			LI_CELL		*pLI = (LI_CELL*) (pData + U32_2PTRDIFF(IndexOffset));

			// Runing on subkeys
			pos = 0;
			while(pos < pLI->KeysCount)
			{
				NK_CELL		*pNKnext = (NK_CELL*) (pData + U32_2PTRDIFF(HHS + pLI->Offset[pos] + LFS));

				// To be on the safe side :)
				if(pNKnext->Signature != ID_NK)
				{
					LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumKey] Incorrect NK cell (li loop).");

					return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
				}

				if(i == dwIndex)
				{
					if(psName)
						*psName = QString(QByteArray(pNKnext->Name, pNKnext->NameLength));

					if(psClass)
					{
						if(pNKnext->OffClassName != 0 && pNKnext->OffClassName != (quint32) -1)
						{
							//
							// NB!
							// I do not test the following because I can not find key belonged to some class.
							//

							const char *pBuf = (const char*)(pData + U32_2PTRDIFF(HHS + pNKnext->OffClassName + LFS));

							*psClass = QString(QByteArray(pBuf, pNKnext->ClassNameLength));
						}
						else
							*psClass = QString();
					}

					return  PRL_ERR_SUCCESS;
				}

				// To next key
				pos++;
				// To next 'global' index
				i++;
			}
		}

		// To next sub child index
		if(pRI)
			ri_pos++;
	}
	while(pRI && ri_pos < pRI->IndexesCount);	// Loop on array of "indexes arrrayes" in case of RI

	/*
	// I (igorl) comment this line as there are o lot of such message in log
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumKey] Index is too large.");
	*/

	return  WinRegistry::REG_INDEX_TOO_LARGE;
}

//
// Function looks for given value among of key values.
//
// Parameters:
//    pNK        [in] - parent key. Exactly among values of this key we will look for given subkey.
//    sValueName [in] - name of value. NB: now we examine only ASCII names.
//
// Return value:
//    -1 if value was not found or its absolute offset in hive otherwise.
//
quint32 CNtBasicFunctions::FindValue(NK_CELL* pNK, const QString& sValueName)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::FindValue]");
	*/

	quint8		*pData = NULL;
	VK_CELL		*pVK = NULL;
	quint32		*pValuesOffset = NULL;
	quint32		i, pos;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::FindValue] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.40:");

		return  (quint32) -1;
	}

	pValuesOffset = (quint32*) (pData + U32_2PTRDIFF(HHS + pNK->OffValuesIndex + LFS));

	for(i = 0, pos = (quint32) -1; i < pNK->ValuesCount; i++)
	{
		pVK = (VK_CELL*) (pData + U32_2PTRDIFF(HHS + pValuesOffset[i] + LFS));

		// To be on the safe size :)
		if(pVK->Signature != ID_VK)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::FindValue] Incorrect VK cell.");
			WRITE_TRACE(DBG_FATAL, "OA00002.41:");

			return  (quint32) -1;
		}

		if(pVK->NameLength == 0 && sValueName.size() == 0)
		{
			pos = i;

			break;
		}

		if(sValueName.size() == pVK->NameLength)
		{
			// Name of value is case insesitive (MSDN said nothing, tested by hands)
			// FIXME: UNICODE support
			if(!Compare(sValueName.toAscii().data(), pVK->Name, pVK->NameLength))
			{
				pos = i;

				break;
			}
		}
	}

	return  pos;
}

//
// Function creates new value in given key.
//
// Parameters:
//    dwNKoffset  [in] - absolute offset of parent key. Exactly among values of this key we will create
//                       new value.
//    sValueName  [in] - name of value. NB: now we examine only ASCII names.
//    dwValueType [in] - type of creted value.
//
// Return value:
//    PRL_ERR_SUCCESS if there are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::CreateValue(quint32 dwNKoffset, const QString& sValueName, quint32 dwValueType)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateValue]");
	*/

	quint8			*pData = NULL;
	NK_CELL			*pNK = NULL;
	VK_CELL			*pVKnew = NULL;
	quint32			VKnewOff;
	quint32			ValueListOffset;
	quint32			pos, size;
	err_status_t	Err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateValue] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.42:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pNK = (NK_CELL*) (pData + U32_2PTRDIFF(dwNKoffset));

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateValue] Incorrect NK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.43:");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	//
	// 1. Check that adding value does not exist (FIXME: this should be
	// implement as separate function)
	//

	pos = FindValue(pNK, sValueName);

	if(!(pos == (quint32) -1))
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateValue] Value with this name is already exist.");
		WRITE_TRACE(DBG_FATAL, "OA00002.44:");

		return  WinRegistry::REG_VALUE_ALREADY_EXIST_ERROR;
	}

	//
	// 2. Allocate necessary space
	//

	// -1 - because one symbol already define in structure NK_CELL
	size = sizeof(VK_CELL) + sValueName.size() - 1;
	Err  = m_pRawData->AllocCell(size, VKnewOff);

	if(PRL_FAILED(Err))
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateValue] Can not allocate space (0x%x bytes) for VK cell (error code: %d).", size, Err);
		WRITE_TRACE(DBG_FATAL, "OA00002.45:\t%x;\t%d", size, Err);

		return  Err;
	}

	size = (pNK->ValuesCount + 1) * sizeof(quint32);
	Err  = m_pRawData->AllocCell(size, ValueListOffset);

	if(PRL_FAILED(Err))
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::CreateValue] Can not allocate space (0x%x bytes) for value list (error code: %d).", size, Err);
		WRITE_TRACE(DBG_FATAL, "OA00002.46:\t%x;\t%d", size, Err);

		m_pRawData->FreeCell(VKnewOff);

		return  Err;
	}

	//
	// 3. Fill new value list
	//

	memcpy(pData + U32_2PTRDIFF(ValueListOffset + LFS), pData + U32_2PTRDIFF(HHS + pNK->OffValuesIndex + LFS), size);

	// Add offset of new value
	*(quint32*) (pData + U32_2PTRDIFF(ValueListOffset + LFS + (pNK->ValuesCount * sizeof(quint32)))) = VKnewOff - HHS;

	// Clear out of old value list
	if(pNK->ValuesCount)
		m_pRawData->FreeCell(pNK->OffValuesIndex + HHS);

	//
	// 4. Fill VK cell of new value
	//

	pVKnew = (VK_CELL*) (pData + U32_2PTRDIFF(VKnewOff + LFS));

	pVKnew->Signature  = ID_VK;
	pVKnew->NameLength = sValueName.size();
	pVKnew->Type       = dwValueType;
	pVKnew->Other1     = 1;

	if(dwValueType == RT_DWORD || dwValueType == RT_DWORD_BIG_ENDIAN)
	{
		pVKnew->DataLength = 0x80000004;
	}
	else
	{
		pVKnew->DataLength = 0x80000001;
	}

	if(sValueName.size())
		strncpy(pVKnew->Name, sValueName.toAscii().data(), sValueName.size());

	//
	// 5. Update NK cell
	//

	pNK->ValuesCount++;
	pNK->OffValuesIndex = ValueListOffset - HHS;

	quint32  dwLen = static_cast<quint32>(sValueName.size()) * 2;

	pNK->Other6 = (pNK->Other6 < dwLen) ? dwLen : pNK->Other6;

	pNK->Other7 = (pNK->Other7 < sizeof(quint32)) ? sizeof(quint32) : pNK->Other7;

	/*
	m_bDirty = true;
	*/

	return  PRL_ERR_SUCCESS;
}

//
// Function deletes given value in given key.
//
// Parameters:
//    dwNKoffset [in] - absolute offset of parent key. Exactly in this key given value
//                      should exist.
//    sValueName [in] - name of deleted value. NB: now we examine only ASCII names.
//
// Return value:
//    PRL_ERR_SUCCESS if there are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::DeleteValue(quint32 dwNKoffset, const QString& sValueName)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteValue]");
	*/

	quint8			*pData = NULL;
	NK_CELL			*pNK = NULL;
	VK_CELL			*pVKdel = NULL;
	quint32			*pValuesOffset = NULL, *pNewValuesOffset = NULL;
	quint32			dwValuesOffset, dwNewValuesOffset, i, j, pos;
	err_status_t	Err;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteValue] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.47:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pNK = (NK_CELL*) (pData + U32_2PTRDIFF(dwNKoffset));

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DelValue] Incorrect NK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.48:");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	if(pNK->ValuesCount == 0)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DelValue] This key has no values.");
		WRITE_TRACE(DBG_FATAL, "OA00002.49:");

		return  WinRegistry::REG_KEY_NO_VALUE_ERROR;
	}

	//
	// 1. Check that deleting value exists (FIXME: this should be
	// implement as separate function)
	//

	pos = FindValue(pNK, sValueName);

	if(pos == (quint32) -1)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DelValue] Value with this name does not exist.");
		WRITE_TRACE(DBG_FATAL, "OA00002.50:");

		return  WinRegistry::REG_VALUE_NOT_EXIST_ERROR;
	}

	pValuesOffset = (quint32*) (pData + U32_2PTRDIFF(HHS + pNK->OffValuesIndex + LFS));

	//
	// 2. Delete VK cell and data
	//

	pVKdel = (VK_CELL*) (pData + U32_2PTRDIFF(HHS + pValuesOffset[pos] + LFS));

	if(pVKdel->Signature != ID_VK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DeleteValue] Incorrect VK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.51:");

		return  WinRegistry::REG_INCORRECT_VK_CELL_ERROR;
	}

	if(!(pVKdel->DataLength & 0x80000000) && pVKdel->OffData)
		m_pRawData->FreeCell(pVKdel->OffData + HHS);

	m_pRawData->FreeCell(HHS + pValuesOffset[pos]);

	//
	// 3. Form new offset array of values and update NK cell
	//

	dwValuesOffset = pNK->OffValuesIndex;

	if(pNK->ValuesCount - 1)
	{
		Err = m_pRawData->AllocCell((pNK->ValuesCount - 1) * sizeof(quint32), dwNewValuesOffset);

		if(PRL_FAILED(Err))
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::DelValue] Can not allocate space (0x%x bytes) for 'offsets array' (error code: %d).", (pNK->ValuesCount - 1) * sizeof(quint32), Err);
			WRITE_TRACE(DBG_FATAL, "OA00002.52:\t%x;\t%d", (pNK->ValuesCount - 1) * sizeof(quint32), Err);

			return  Err;
		}

		// Now copy over, omitting deleted entry
		pNewValuesOffset = (quint32*) (pData + U32_2PTRDIFF(dwNewValuesOffset + LFS));

		for(i = 0, j = 0; i < pNK->ValuesCount; i++)
		{
			if(i != pos)
			{
				pNewValuesOffset[j] = pValuesOffset[i];

				j++;
			}
		}

		pNK->OffValuesIndex = dwNewValuesOffset - HHS;
		pNK->ValuesCount--;
	}
	else
	{
		pNK->OffValuesIndex = (quint32) -1;
		pNK->ValuesCount--;
	}

	return  m_pRawData->FreeCell(dwValuesOffset + HHS);
}

//
// Function returns data of given value in given key.
//
// Parameters:
//    dwNKoffset     [in]     - absolute offset of parent key. Exactly in this key given value should exist.
//    sValueName     [in]     - name of value. NB: now we examine only ASCII names.
//    dwValueType    [in]     - type of value.
//    pBuf           [out]    - data of value will be copied in this buffer.
//    dwBufSize      [in]     - size of buffer.
//    dwRealDataSize [out]    - size of data.
//    pdwValueType   [in/out] - real type of value. This parameter may be NULL.
//
// FIXME : This function must have [out,optional] parameter dwValueType,
// pdwValueType parameter is used as work around.
//
// Return value:
//    PRL_ERR_SUCCESS if there are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::GetValueData(
									quint32			dwNKoffset,
									const QString&	sValueName,
									quint32			dwValueType,
									quint8*			pBuf,
									quint32			dwBufSize,
									quint32&		dwRealDataSize,
									quint32*		pdwValueType)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::GetValueData]");
	*/

	quint8		*pData = NULL;
	NK_CELL		*pNK = NULL;
	VK_CELL		*pVK = NULL;
	quint32		*pValuesOffset = NULL;
	quint32		pos;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::GetValueData] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.53:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pNK = (NK_CELL*) (pData + dwNKoffset);

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::GetValueData] Incorrect NK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.54:");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	if(pNK->ValuesCount == 0)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::GetValueData] This key has no values.");
		WRITE_TRACE(DBG_FATAL, "OA00002.55:");

		return  WinRegistry::REG_KEY_NO_VALUE_ERROR;
	}

	//
	// 1. Check that deleting value exists (FIXME: this should be
	// implement as separate function)
	//

	pos = FindValue(pNK, sValueName);

	if(pos == (quint32) -1)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::GetValueData] Value with this name does not exist.");
		WRITE_TRACE(DBG_FATAL, "OA00002.56:");

		return  WinRegistry::REG_VALUE_NOT_EXIST_ERROR;
	}

	pValuesOffset = (quint32*) (pData + U32_2PTRDIFF(HHS + pNK->OffValuesIndex + LFS));

	pVK = (VK_CELL*) (pData + U32_2PTRDIFF(HHS + pValuesOffset[pos] + LFS));

	if(pVK->Signature != ID_VK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::GetValueData] Incorrect VK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.57:");

		return  WinRegistry::REG_INCORRECT_VK_CELL_ERROR;
	}

	// Save value type as out param
	if ( pdwValueType )
	{
		*pdwValueType = pVK->Type;
	}

	if(pVK->DataLength == 0)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::GetValueData] This value is empty (DataLenght is zero).");
		WRITE_TRACE(DBG_FATAL, "OA00002.58:");

		return  WinRegistry::REG_VALUE_IS_EMPTY_ERROR;
	}

	if(pVK->DataLength == 0x80000000)
	{
		dwRealDataSize = sizeof(pVK->Type);

		// To be on the safe side :)
		if(dwRealDataSize > dwBufSize)
		{
			if (dwBufSize != 0)
				WRITE_TRACE(DBG_FATAL, "OA00002.59: buffer to small %x; %x", dwRealDataSize, dwBufSize);

			return   WinRegistry::REG_BUFFER_TOO_SMALL_ERROR;
		}

		// Special inline case (len = 0x80000000): data (4 bytes?) in type field
		memcpy(pBuf, (quint8*) &(pVK->Type), sizeof(pVK->Type));

		return  PRL_ERR_SUCCESS;
	}

	if(dwValueType && pVK->Type && pVK->Type != dwValueType)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::GetValueData] Data type is incorrect.");
		WRITE_TRACE(DBG_FATAL, "OA00002.60:");

		return  WinRegistry::REG_INCORRECT_DATA_TYPE_ERROR;
	}

	// Negative len is inline, return ptr to offset-field which in this case
	// contains the data itself
	// FIXME: this may be incorrect, need to test for REG_BINARY and DataLength < 4
	if(pVK->DataLength & 0x80000000)
	{
		dwRealDataSize = sizeof(pVK->OffData);

		// To be on the safe side :)
		if(dwRealDataSize > dwBufSize)
		{
			if (dwBufSize != 0)
				WRITE_TRACE(DBG_FATAL, "OA00002.61: buffer to small %x; %x", dwRealDataSize, dwBufSize);

			return   WinRegistry::REG_KEY_NO_VALUE_ERROR;
		}

		memcpy(pBuf, (quint8*) &(pVK->OffData), sizeof(pVK->OffData));

		return  PRL_ERR_SUCCESS;
	}

	if(pVK->DataLength > dwBufSize)
	{
		if (dwBufSize != 0)
			WRITE_TRACE(DBG_FATAL, "OA00002.62: buffer to small %x; %x", pVK->DataLength, dwBufSize);

		dwRealDataSize = pVK->DataLength;

		return   WinRegistry::REG_BUFFER_TOO_SMALL_ERROR;
	}

	dwRealDataSize = pVK->DataLength;

	memcpy(pBuf, (quint8*) (pData + U32_2PTRDIFF(HHS + pVK->OffData + LFS)), pVK->DataLength);

	return  PRL_ERR_SUCCESS;
}

//
// Function sets new data for given value in given key.
//
// Parameters:
//    dwNKoffset     [in]  - absolute offset of parent key. Exactly in this key given value should exist.
//    sValueName     [in]  - name of value. NB: now we examine only ASCII names.
//    dwValueType    [in]  - type of value.
//    pBuf           [in]  - data from this buffer will be seted to the value.
//    dwBufSize      [in]  - size of data.
//
// Return value:
//    PRL_ERR_SUCCESS if there are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::SetValueData(quint32 dwNKoffset, const QString& sValueName, quint32 dwValueType, quint8* pBuf, quint32 dwBufSize)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::SetValueData]");
	*/

	quint8		*pData = NULL;
	NK_CELL		*pNK = NULL;
	VK_CELL		*pVK = NULL;
	quint32		*pValuesOffset = NULL;
	quint32		pos, Len;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::SetValueData] Storage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00002.63:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pNK = (NK_CELL*) (pData + U32_2PTRDIFF(dwNKoffset));

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::SetValueData] Incorrect NK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.64:");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	if(pNK->ValuesCount == 0)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::SetValueData] This key has no values.");
		WRITE_TRACE(DBG_FATAL, "OA00002.65:");

		return  WinRegistry::REG_VALUE_NOT_EXIST_ERROR; // WinRegistry::REG_KEY_HAS_NO_VALUE
	}

	//
	// 1. Check that deleting value exists (FIXME: this should be
	// implement as separate function)
	//

	pos = FindValue(pNK, sValueName);

	if(pos == (quint32) -1)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::SetValueData] Value with this name does not exist.");
		WRITE_TRACE(DBG_FATAL, "OA00002.66:");

		return  WinRegistry::REG_VALUE_NOT_EXIST_ERROR;
	}

	pValuesOffset = (quint32*) (pData + U32_2PTRDIFF(HHS + pNK->OffValuesIndex + LFS));

	pVK = (VK_CELL*) (pData + U32_2PTRDIFF(HHS + pValuesOffset[pos] + LFS));

	if(pVK->Signature != ID_VK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::SetValueData] Incorrect VK cell.");
		WRITE_TRACE(DBG_FATAL, "OA00002.67:");

		return  WinRegistry::REG_INCORRECT_VK_CELL_ERROR;
	}

	Len = pVK->DataLength & 0x7FFFFFFF;

	if(pVK->DataLength == 0x80000000)
	{
		// Special inline case, return size of 4 (dword)
		Len = 4;
	}

	if(dwBufSize != Len)
	{
		quint32			dwDataOffset;
		err_status_t	Err;

		Err = m_pRawData->AllocCell(dwBufSize, dwDataOffset);

		if(PRL_FAILED(Err))
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::SetValueData] Can not allocate space (0x%x bytes) for data (error code: %d).", dwBufSize, Err);
			WRITE_TRACE(DBG_FATAL, "OA00002.68:\t%x;\t%d", dwBufSize, Err);

			return  Err;
		}

		if(pVK->OffData != 0 && pVK->OffData != quint32(-1))
			m_pRawData->FreeCell(pVK->OffData + HHS);

		pVK->OffData    = dwDataOffset - HHS;
		pVK->DataLength = dwBufSize;
	}

	if(pVK->DataLength == 0x80000000)
	{
		// Special inline case (len = 0x80000000): data (4 bytes?) in type field

		memset(&(pVK->Type), 0, sizeof(pVK->Type));
		memcpy(&(pVK->Type), pBuf, dwBufSize);

		pNK->Other7 = (pNK->Other7 < sizeof(pVK->Type)) ? sizeof(pVK->Type) : pNK->Other7;

		return  PRL_ERR_SUCCESS;
	}

	if(dwValueType && pVK->Type && pVK->Type != dwValueType)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::SetValueData] Data type is incorrect.");
		WRITE_TRACE(DBG_FATAL, "OA00002.69:");

		return  WinRegistry::REG_INCORRECT_DATA_TYPE_ERROR;
	}

	// Negative len is inline, return ptr to offset-field which in this case
	// contains the data itself
	if(pVK->DataLength & 0x80000000)
	{
		memset(&(pVK->OffData), 0, sizeof(pVK->OffData));
		memcpy(&(pVK->OffData), pBuf, dwBufSize);

		pNK->Other7 = (pNK->Other7 < sizeof(pVK->OffData)) ? sizeof(pVK->OffData) : pNK->Other7;

		return  PRL_ERR_SUCCESS;
	}

	memcpy((pData + U32_2PTRDIFF(HHS + pVK->OffData + LFS)), pBuf, dwBufSize);

	pNK->Other7 = (pNK->Other7 < dwBufSize) ? dwBufSize : pNK->Other7;

	return  PRL_ERR_SUCCESS;
}

//
// Function enumerates the values for the specified open registry key.
//
// Parameters:
//
//    dwOffset   [in]  - absolute offset of specified key
//    dwIndex    [in]  - index of the value to be retrieved
//    psValue    [out] - pointer to a QString that receives the name of the value (optinal)
//    pdwType    [out] - pointer to a variable that receives a code indicating the type of
//                       data stored in the specified value (optinal)
//    pBuf       [out] - not support yet, must be NULL now
//    pdwBufSize [out] - not support yet, must be NULL now
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtBasicFunctions::EnumValue(quint32 dwOffset, quint32 dwIndex, QString* psValue, quint32* pdwType, quint8* pBuf, quint32* pdwBufSize)
{
	/*
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumValue]");
	*/

	quint8		*pData = NULL;
	NK_CELL		*pNK = NULL;
	VK_CELL		*pVK = NULL;
	quint32		*pValuesOffset = NULL;
	quint32		i;

	pData = m_pRawData ? m_pRawData->GetRawData() : NULL;

	if(!pData)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumValue] Storage is NULL.");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	if(dwOffset == quint32(-1))
		// Offset was not determine, lets go from root key
		dwOffset = HHS + m_pRawData->GetRootCellOffset() + LFS;

	pNK = (NK_CELL*) (pData + U32_2PTRDIFF(dwOffset));

	if(pNK->Signature != ID_NK)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumValue] Incorrect NK cell.");

		return  WinRegistry::REG_INCORRECT_NK_CELL_ERROR;
	}

	if(pBuf || pdwBufSize)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumValue] Unsupported features of this function.");

		return  WinRegistry::REG_UNSUPPORTED_ERROR;
	}

	//
	// Loop on values (looking for specified index).
	//
	// P.S.
	// The same loop already exists in FindValue.
	//

	pValuesOffset = (quint32*) (pData + U32_2PTRDIFF(HHS + pNK->OffValuesIndex + LFS));

	for(i = 0; i < pNK->ValuesCount; i++)
	{
		pVK = (VK_CELL*) (pData + U32_2PTRDIFF(HHS + pValuesOffset[i] + LFS));

		// To be on the safe side :)
		if(pVK->Signature != ID_VK)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumValue] Incorrect VK cell (loop).");

			return  WinRegistry::REG_INCORRECT_VK_CELL_ERROR;
		}

		if(i == dwIndex)
		{
			if(psValue)
				*psValue = QString(QByteArray(pVK->Name, pVK->NameLength));

			if(pdwType)
				*pdwType = pVK->Type;

			return  PRL_ERR_SUCCESS;
		}
	}

	/*
	// I (igorl) comment this line as there are o lot of such message in log
	LOG_MESSAGE(DBG_FATAL, "[CNtBasicFunctions::EnumValue] Index is too large.");
	*/

	return  WinRegistry::REG_INDEX_TOO_LARGE;
}

