/////////////////////////////////////////////////////////////////////////////
///
/// @file NtBasicFunctions.h
///
/// This module implements "low-level" functions for working with keys and
/// values of "NT" registry hive.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#ifndef __NT__BASIC__FUNCTIONS__H__
#define __NT__BASIC__FUNCTIONS__H__

#include "IBasicFunctions.h"
#include "RegTypes.h"
#include "NtRawData.h"

///////////////////////////////////////////////////////////////////////////////
// class CNtBasicFunctions

class CNtBasicFunctions : public IBasicFunctions
{
	CNtRawData	*m_pRawData;

#include PRE_PACK
	struct NK_CELL
	{
		quint16	Signature;
		// 0x2c - root, 0x20 - otherwise
		// 0x10 - symbolic link
		// 0x20 - key name does not use UNICODE
		quint16	Type;
		quint8	TimeStamp[8];
		// Padding (0x00)
		quint32	Other1;
		quint32	ParentOffset;
		quint32	SubKeysCount;
		// Padding (0x00)
		quint32	Other2;
		quint32	OffSubKeysIndex;
		// 0xFFFFFFFF almost always
		quint32	Other3;
		quint32	ValuesCount;
		quint32	OffValuesIndex;
		quint32	OffSk;
		quint32	OffClassName;
		// Length of longest sub key name * 2
		// P.S.
		// We do not update this field during deleting of sub key. It
		// seems to us that "regedit" does the same.
		quint32	Other4;
		// Length of longest class name of sub key
		// P.S.
		// We do not update this field during deleting of sub key. It
		// seems to us that "regedit" does the same.
		quint32	Other5;
		// Length of longest value name * 2
		// P.S.
		// We do not update this field during deleting of value. It
		// seems to us that "regedit" does the same.
		quint32	Other6;
		// Size of biggest value data
		// P.S.
		// We do not update this field during deleting of value. It
		// seems to us that "regedit" does the same.
		quint32	Other7;
		quint32	Other8;
		quint16	NameLength;
		quint16	ClassNameLength;
		char	Name[1];
	}PACKED;
#include POST_PACK

#include PRE_PACK
	struct LF_LH_CELL
	{
		quint16	Signature;
		quint16	KeysCount;

		struct LF_LH_HASH
		{
			quint32	Offset;
			char	Name[4];	// for 'lh' index cell this field is simply quint32
		}PACKED Hash[1];

	}PACKED;
#include POST_PACK

#include PRE_PACK
	struct LI_CELL
	{
		quint16	Signature;
		quint16	KeysCount;
		quint32	Offset[1];
	}PACKED;
#include POST_PACK

#include PRE_PACK
	struct RI_CELL
	{
		quint16	Signature;
		quint16	IndexesCount;
		quint32	Offset[1];
	}PACKED;
#include POST_PACK

#include PRE_PACK
	struct SK_CELL
	{
		quint16	Signature;
		quint16	Other1;
		quint32	OffPrevSk;
		quint32	OffNextSk;
		quint32	UsageCount;
		quint32	Length;
		quint8	Data[1];
	}PACKED;
#include POST_PACK

#include PRE_PACK
	struct VK_CELL
	{
		quint16 Signature;
		quint16 NameLength;
		quint32 DataLength;
		quint32 OffData;
		quint32 Type;
		quint16 Other1;
		quint16 Other2;
		char    Name[1];
	}PACKED;
#include POST_PACK

	// Signature of key cell (0x6b6e = 'nk')
	const static quint16	ID_NK = 0x6b6e;
	// Signature of child index cell (0x666c = 'lf', Windows 2000)
	const static quint16	ID_LF = 0x666c;
	// Signature of child index cell (0x686c = 'lh', Windows XP)
	const static quint16	ID_LH = 0x686c;
	// Signature of child index cell if quantity of child is big (0x696c = 'li')
	const static quint16	ID_LI = 0x696c;
	// Signature of root index of child index cell (0x6972 = 'ri')
	const static quint16	ID_RI = 0x6972;
	// Signature of security cell (0x6b73 = 'sk')
	const static quint16	ID_SK = 0x6b73;
	// Signature of value cell (0x6b76 = 'vk')
	const static quint16	ID_VK = 0x6b76;

	// This is maximum number of keys described by "lf" cell.
	// This constant is correct for hives of Windows 2000 ("lf" index cell):
	// I have checked this fact "by hand". But I could exceed this value for
	// hives of Windows XP ("lh" index cell).
	const static quint16	MAX_LF_LH_COUNT = 0x1FB;

	// This is maximum number of keys described by "li" cell.
	// Paul Lee contends that this value must be 0x3F5. But when I have made
	// some experiments with hives of Windows 2000 I saw that this value must
	// be 0x3F4.
	const static quint16	MAX_LI_COUNT = 0x3F4;


	/**
	 * Compares the first iSize bytes of pLeft and pRight without regard to
	 * the case of letters in the two buffers. The function converts all
	 * lowercase characters into uppercase (!) and then performs the comparison.
	 *
	 * @author igorl
	 *
	 * @param[in] pLeft is the pointer to first buffer.
	 * @param[in] pRight is the pointer to second buffer.
	 * @param[in] iSize is the number of characters.
	 *
	 * @return -1 if pLeft is less than pRight, 0 if pLeft is identical to pRight,
	 * 1 if pLeft is greater than pRight.
	 */
	int Compare(const char* pLeft, const char* pRight, int iSize) const;


	quint32 FindSubKey(NK_CELL* pNK, const char* szSubKeyName, quint16& pos, quint16& ri_pos);
	quint32 FindValue(NK_CELL* pNK, const QString& sValueName);


	/**
	 * Adds new Lf/Lh index cell to existed Ri cell, i.e. calculates size of
	 * new Ri cell, allocates needed space, fills new Ri cell and frees old
	 * one.
	 *
	 * @author igorl
	 *
	 * @param[in] dwRiOffOld is relative offset of existed Ri cell.
	 * @param[in] dwLfLhOff is absolute offset of adding Lf/Lh index cell.
	 * @param[out] dwRiOffNew is absolute offset of new Ri cell.
	 *
	 * @return  R_SUCCESS if there are no error or detailed error code
	 * otherwise.
	 */
	err_status_t AddLfLhToRi(
						quint32		dwRiOffOld,
						quint32		dwLfLhOff,
						quint32&	dwRiOffNew);


	/**
	 * We have full Li index cell (addressed by dwLIoff) and new adding key
	 * (with specified name and offset) is hit to this Li cell. We should find
	 * corresponded place for this new key, place it there, shift other keys
	 * and receive name and offset of pushed old key.
	 * NB!
	 * This function checks fullness of Li index cell.
	 *
	 * @author igorl
	 *
	 * @param[in/out] sKeyName is the name of new adding key on input and
	 * the name of pushed old key on output.
	 * @param[in/out] dwNKoff is the offset of new adding key on input and
	 * the offset of pushed old key on output. Both offsets are absolute.
	 * @param[in] dwLIoff is the relative offset of editing Li cell.
	 *
	 * @return R_SUCCESS if there are no error or detailed error code
	 * otherwise.
	 */
	err_status_t Swap(
					QString&	sKeyName,
					quint32&	dwNKoff,
					quint32		dwLIoff);


	err_status_t CreateLfLh(const char* szKeyName, quint32 NKnewOff, quint32& IndexOffset);
	err_status_t UpdateLfLh(const char* szKeyName, quint32 NKnewOff, quint32 IndexOffsetOld, quint32& IndexOffsetNew);
	err_status_t UpdateLi(const char* szKeyName, quint32 NKnewOff, quint32 IndexOffsetOld, quint32& IndexOffsetNew);


	/**
	 * Updates existing Ri index cell.
	 *
	 * @author igorl
	 *
	 * @param[in] szKeyName is the name of new adding key. Now only ASCII names
	 * are supported.
	 * @param[in] dwNKnewOff is the absolute offset of new adding key.
	 * @param[in] dwRiIndexOffsetOld is the relative offset of Ri index cell of
	 * parent key.
	 * @param[out] dwRiIndexOffsetNew is the absolute offset of updated Ri cell.
	 *
	 * @return R_SUCCESS if the are no error or detailed error code otherwise.
	 */
	err_status_t UpdateRi(
					const char*		szKeyName,
					quint32			dwNKnewOff,
					quint32			dwRiIndexOffsetOld,
					quint32&		dwRiIndexOffsetNew);


	/**
	 * Creates "li" index cell based on "lf/lh" index cell. Previous
	 * "lf/lh" index cell will be freed in case of success creation
	 * of new index cell.
	 * NB!
	 * You need in this function when UpdateLfLh returns
	 * IDE_REG_INDEXES_ARRAY_FULL error.
	 *
	 * @author igorl
	 *
	 * @param[in] szKeyName is the name of new created key.
	 * This name is used for determination of correct placement
	 * the created key in index cell.
	 * @param[in] dwNKnewOff is the offset of new created key.
	 * The offset must be absolute.
	 * @param[in] dwIndexOffsetOld is the offset of previous
	 * index cell. The offset must be relative.
	 * @param[out] dwIndexOffsetNew is the offset of new created
	 * "li" index cell. This offset is absolute.
	 *
	 * @return R_SUCCESS if there are no error or detailed error
	 * code otherwise.
	 */
	err_status_t ConvertLfLhToLi(
						const char*		szKeyName,
						quint32			dwNKnewOff,
						quint32			dwIndexOffsetOld,
						quint32&		dwIndexOffsetNew);


	/**
	 * Creates Ri index cell based on Li index cell. Previous Li
	 * index cell will be re-used.
	 * NB!
	 * You need in this function when UpdateLi returns
	 * IDE_REG_INDEXES_ARRAY_FULL error.
	 *
	 * @author igorl
	 *
	 * @param[in] szKeyName is the name of new created key.
	 * This name is used for determination of correct placement
	 * the created key in index cell.
	 * @param[in] dwNKnewOff is the offset of new created key.
	 * The offset must be absolute.
	 * @param[in] dwLiOff is the offset of previous index cell
	 * The offset must be relative.
	 * @param[out] dwRiOff is the offset of new created Ri index
	 * cell. This offset is absolute.
	 *
	 * @return R_SUCCESS if there are no error or detailed error
	 * code otherwise.
	 */
	err_status_t ConvertLiToRi(
							const char*		szKeyName,
							quint32			dwNKnewOff,
							quint32			dwLiOff,
							quint32&		dwRiOff);


	err_status_t RebuildIndex(const char* szKeyName, quint32 NKnewOff, quint32 IndexOffsetOld, quint32& IndexOffsetNew);

public:
	CNtBasicFunctions(QByteArray *pData, int iDelta)
	{
		 m_pRawData = new CNtRawData(pData, iDelta);
	}
	virtual ~CNtBasicFunctions()
	{
		if(m_pRawData)
			delete  m_pRawData;
	}

	virtual err_status_t OpenKey(const QString& sKeyName, quint32& dwKeyOffset, quint32 dwParentOffset = (quint32) -1);
	virtual err_status_t CreateKey(const QString& sKeyName, quint32& dwKeyOffset, quint32 dwOffset = (quint32) -1);
	virtual err_status_t DeleteKey(const QString& sKeyName, quint32 dwOffset = (quint32) -1);
	virtual err_status_t QueryInfoKey(quint32 dwOffset, QString* psClassName, quint32* pcSubKeys, quint16* pMaxSubKeyLen, quint16* pMaxClassLen, quint32* pcValues, quint16* pMaxValueLen, quint32* pMaxValueDataLen);
	virtual err_status_t EnumKey(quint32 dwOffset, quint32 dwIndex, QString* psName, QString* psClass);

	virtual err_status_t CreateValue(quint32 dwNKoffset, const QString& sValueName, quint32 dwValueType);
	virtual err_status_t DeleteValue(quint32 dwNKoffset, const QString& sValueName);

	// FIXME : This function must have [out,optional] parameter dwValueType,
	// pdwValueType parameter is used as work around.
	virtual err_status_t GetValueData(
								quint32			dwNKoffset,
								const QString&	sValueName,
								quint32			dwValueType,
								quint8*			pBuf,
								quint32			dwBufSize,
								quint32&		dwRealDataSize,
								quint32*		pdwValueType = NULL);

	virtual err_status_t SetValueData(quint32 dwNKoffset, const QString& sValueName, quint32 dwValueType, quint8* pBuf, quint32 dwBufSize);
	virtual err_status_t EnumValue(quint32 dwOffset, quint32 dwIndex, QString* psValue, quint32* pdwType, quint8* pData, quint32* pdwDataSize);

	virtual void Free()
	{
		delete  this;
	}
};

#endif // __NT__BASIC__FUNCTIONS__H__
