/////////////////////////////////////////////////////////////////////////////
///
/// @file NtRawData.cpp
///
/// This module implements functions for working with raw data of "NT"
/// registry hive. These functions are identification, receiving of different
/// version information, management of raw space.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#include "NtRawData.h"

///////////////////////////////////////////////////////////////////////////////
// class CNtRawData

//
// Function calculates check summa of hive header.
//
// Return value:
//    -1 if check summa was not calculated or check summa itself otherwise.
//
qint32 CNtRawData::CalcCheckSum(void)
{
	qint32	*pData,	dwSum;

	if(!m_pStorage)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::CalcCheckSum] m_pStorage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00003.01:");

		return	-1;
	}

	//
	// Check sum is calculated by operation "xor" under all dwords from
	// hive header.
	//

	pData = (qint32*) m_pStorage->data();

	dwSum = 0x0;
	for(unsigned i = 0; i < sizeof(HIVE_HEADER) / sizeof(qint32) - 1; i++)
		dwSum ^= pData[i];

	return  dwSum;
}

//
// Function looks for free cell of given size.
//
// Parameters:
//    dwSize   [in]  - free cell of such size will be looked for.
//    dwOffset [out] - absolute offset of found cell with given size. If free cell with such size
//                     is not found this parameter will be equal -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtRawData::FindFreeCell(qint32 dwSize, quint32& dwOffset)
{
	quint8		*pRawData = NULL;
	BIN_HEADER	*pHBin = NULL;
	quint32		dwOffNextBin, dwOffBin, dwOffCell;
	qint32		dwCellSize;

	dwOffset = (quint32) -1;

	if(!m_pStorage)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::FindFreeCell] m_pStorage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00003.02:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pRawData = (quint8*) m_pStorage->data();
	// dwOffBin - absolute offset of begining of current bin
	dwOffBin = HIVE_HDR_SIZE;

	while(dwOffBin < (quint32) m_pStorage->size())
	{
		//
		// Bin loop
		//

		pHBin = (BIN_HEADER*) (pRawData + dwOffBin);

		if(pHBin->Signature != ID_HBIN)
		{
			//
			// This is not a HBin, assuming (P. Nordahl-Hagen) file contains garbage at
			// end.  I simply cut off hive up to its real size.  Istead dwOffBin we may
			// write pHdr->Length + HIVE_HDR_SIZE.
			//

			LOG_MESSAGE(DBG_FATAL, "[CNtRawData::FindFreeCell] Incorrect bin (signature 0x%x, offset 0x%x, file size 0x%x).", pHBin->Signature, dwOffBin, m_pStorage->size());
			WRITE_TRACE(DBG_FATAL, "OA00003.03:\t%x;\t%x;\t%x", pHBin->Signature, dwOffBin, m_pStorage->size());

			if((int) dwOffBin < m_pStorage->capacity())
				m_pStorage->resize(dwOffBin);
			else
			{
				LOG_MESSAGE(DBG_FATAL, "[CNtRawData::FindFreeCell] Can not reallocate memory for hive.");
				WRITE_TRACE(DBG_FATAL, "OA00003.11");

				return	WinRegistry::REG_REALLOCATE_ERROR;
			}

			return  WinRegistry::REG_NO_FREE_CELL_ERROR;	// WinRegistry::INCORECT_BIN
		}

		// To be on the safe side
		if(pHBin->Size == 0)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtRawData::FindFreeCell] Size of bin is equal 0 (signature 0x%x, offset 0x%x, file size 0x%x).", pHBin->Signature, dwOffBin, m_pStorage->size());
			WRITE_TRACE(DBG_FATAL, "OA00003.04:\t%x;\t%x;\t%x", pHBin->Signature, dwOffBin, m_pStorage->size());

			return  WinRegistry::REG_INCORECT_BIN_ERROR;
		}

		// dwOffNextBin - absolute offset of next bin (i.e. end of current)
		dwOffNextBin = dwOffBin + pHBin->Size;
		// dwOffCell    - absolute offset of current analyzable cell
		dwOffCell    = dwOffBin + sizeof(BIN_HEADER);

		while(dwOffCell < dwOffNextBin)
		{
			//
			// Cell loop
			//

			dwCellSize = *((qint32*) (pRawData + dwOffCell));

			if(dwCellSize > 0 && dwCellSize >= dwSize)
			{
				//
				// Find first proper cell
				//

				dwOffset = dwOffCell;

				return  PRL_ERR_SUCCESS;
			}

			// Pass to next cell
			if(dwCellSize < 0)
				dwOffCell += -dwCellSize;
			else
			{
				if(dwCellSize > 0)
					dwOffCell += dwCellSize;
				else
				{
					LOG_MESSAGE(DBG_FATAL, "[CNtRawData::FindFreeCell] Cell with zero size was found (cell offset 0x%x).", dwOffCell);
					WRITE_TRACE(DBG_FATAL, "OA00003.05:\t%x", dwOffCell);

					return WinRegistry::REG_ZERO_CELL_SIZE_ERROR;
				}
			}
		}

		// Pass to next bin
		dwOffBin = dwOffNextBin;
	}

	return  WinRegistry::REG_NO_FREE_CELL_ERROR;
}

//
// Function adds new bin to hive.
//
// Parameters:
//    dwSize   [in]  - size of added bin (this size may be corrected as bin size should be multiplied
//                     of 4k).
//    dwOffset [out] - absolute offset of free space (!) in added bin. If bin can not be added this
//                     parameter will be equal -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtRawData::AddBin(qint32 dwSize, quint32& dwOffset)
{
	quint8			*pRawData = NULL;
	HIVE_HEADER		*pHdr = NULL;
	BIN_HEADER		*pHBin = NULL;
	qint32			*pCell = NULL;

	dwOffset = (quint32) -1;

	if(!m_pStorage)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::AddBin] m_pStorage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00003.06:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pHdr   = (HIVE_HEADER*) m_pStorage->data();

	// Bin size should be multiplied of 4k
	/*
	dwSize = (dwSize + 0xFFF) & ~0xFFF;
	*/
	// corrected by mityak@
	dwSize = (dwSize + sizeof(BIN_HEADER) + 0xFFF) & ~0xFFF;

	// Allocate space
	if(m_pStorage->size() + dwSize < m_pStorage->capacity())
		m_pStorage->resize(m_pStorage->size() + dwSize);
	else
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::AddBin] Can not reallocate memory for hive.");
		WRITE_TRACE(DBG_FATAL, "OA00003.11");

		return  WinRegistry::REG_REALLOCATE_ERROR;
	}

	// pRawData points to allocated space
	pRawData = (quint8*) (m_pStorage->data() + m_pStorage->size() - dwSize);

	memset(pRawData, 0, dwSize);

	// Fill some fields in bin header
	pHBin = (BIN_HEADER*) pRawData;

	pHBin -> Signature  = ID_HBIN;
	pHBin -> FileOffset = pHdr->Length;
	pHBin -> Size       = dwSize;
	pHBin -> Spare      = dwSize;

	// Fill some fields in cell
	pCell = (qint32*) (pRawData + sizeof(BIN_HEADER));

	pCell[0] = dwSize - sizeof(BIN_HEADER);
	// FIXME: According to Microsoft presentation (p. 12) free cells in the same
	// 'size class' have (?) to link together.
	pCell[1] = -1;

	// Correct some fields in hive header
	pHdr->Length   = pHdr->Length + dwSize;
	pHdr->CheckSum = CalcCheckSum();

	// Return offset to cell
	dwOffset = m_pStorage->size() - dwSize + sizeof(BIN_HEADER);

	return  PRL_ERR_SUCCESS;
}

//
// Function returns offset of free cell with given size. This cell will be found amoung
// existing free cells or will be added to hive if there are no free cell of such size.
//
// Parameters:
//    dwSize   [in]  - cell of such size is needed.
//    dwOffset [out] - absolute offset of found/added cell with given size. If cell with
//                     such size can not be found/added this parameter will be equal -1.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtRawData::AllocCell(qint32 dwSize, quint32& dwOffset)
{
	quint8			*pRawData = NULL;
	qint32			dwOldSize, dwTrailSize;
	err_status_t	errCode;

	dwOffset = (quint32) -1;

	if(!m_pStorage)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::AllocCell] m_pStorage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00003.07:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pRawData = (quint8*) m_pStorage->data();

	// NB!
	// Add size of size field of cell
	dwSize = dwSize + LEN_FIELD_SIZE;

	// Size should be rounded at 8 bytes boundary
	if(dwSize & 7)
		dwSize += (8 - (dwSize & 7));

	// Try to find free cell
	errCode = FindFreeCell(dwSize, dwOffset);

	if(PRL_FAILED(errCode))
	{
		// Search was not successful, check error code

		if(errCode == WinRegistry::REG_NO_FREE_CELL_ERROR)
		{
			if(PRL_FAILED(AddBin(dwSize, dwOffset)))
			{
				LOG_MESSAGE(DBG_FATAL, "[CNtRawData::AllocCell] Can not add bin (something very strage).");
				WRITE_TRACE(DBG_FATAL, "OA00003.08:");

				return  WinRegistry::REG_ALLOCATE_SPACE_ERROR;
			}
		}
		else
			return  errCode;
	}

	// Get the space
	dwOldSize   = *((qint32*) (pRawData + dwOffset));
	dwTrailSize = dwOldSize - dwSize;

	// FIXME: This part of code was spied from P. Nordahl-Hagen. It seems to me
	// a little strange. IgorL@

	if(dwTrailSize == 4)
	{
		dwTrailSize  = 0;
		dwSize      += 4;
	}

	if(dwTrailSize & 7)
	{
		// Trail must be 8 aligned
		dwTrailSize -= (8 - (dwTrailSize & 7));
		dwSize      += (8 - (dwTrailSize & 7));		// ???
	}

	// Now change pointers on this to reflect new size
	*((qint32*) (pRawData + dwOffset)) = -dwSize;

	// To be on the safe size (4 = sizeof(size of data))
	memset((void*) (pRawData + dwOffset + LEN_FIELD_SIZE), 0, dwSize - LEN_FIELD_SIZE);

	if(dwTrailSize)
	{
		*((qint32*)(pRawData + dwOffset + dwSize)) = dwTrailSize;
		// FIXME: According to Microsoft presentation (p. 12) free cells in the same
		// 'size class' have (?) to link together.
		*((qint32*)(pRawData + dwOffset + dwSize + LEN_FIELD_SIZE)) = -1;

		memset((void*) (pRawData + dwOffset + dwSize + 2 * LEN_FIELD_SIZE), 0, dwTrailSize - 2 * LEN_FIELD_SIZE);
	}

	return  PRL_ERR_SUCCESS;
}

//
// Function frees cell with given offset.
//
// Parameters:
//    dwOffset [in] - absolute offset of given cell (!) i.e. first field of cell is its size.
//
// Return value:
//    PRL_ERR_SUCCESS if the are no error or detailed error code otherwise.
//
err_status_t CNtRawData::FreeCell(quint32 dwOffset)
{
	qint32	dwSize;
	qint32	*pCell;

	if(!m_pStorage)
	{
		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::FreeCell] m_pStorage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00003.09:");

		return  WinRegistry::REG_EMPTY_STORAGE_ERROR;
	}

	pCell  = (qint32*) (m_pStorage->data() + dwOffset);

	dwSize = pCell[0];

	if(dwSize >= 0)
	{
		// something strange: we are trying to free already freed cell

		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::FreeCell] Trying to free already freed cell (offset 0x%x, size 0x%x).", dwOffset, dwSize);
		WRITE_TRACE(DBG_FATAL, "OA00003.10:\t%x;\t%x", dwOffset, dwSize);

		return	WinRegistry::REG_FREE_CELL_ERROR;
	}

	dwSize = -dwSize;

	memset(pCell, 0, dwSize);

	pCell[0] = dwSize;
	// FIXME: According to Microsoft presentation (p. 12) free cells in the same
	// 'size class' have (?) to link together.
	pCell[1] = -1;

	return  PRL_ERR_SUCCESS;
}
