/////////////////////////////////////////////////////////////////////////////
///
/// @file NtRawData.h
///
/// This module implements functions for working with raw data of "NT"
/// registry hive. These functions are identification, receiving of different
/// version information, management of raw space.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#ifndef	__NT__RAW__DATA__H__
#define	__NT__RAW__DATA__H__

#include "Pack.h"
#include <QByteArray>
#include "WinRegistryErrors.h"
#include <prlcommon/Logging/Logging.h>

// Size of hive header
#define	HIVE_HDR_SIZE	0x1000
// Length (in bytes) of cell size field
#define LEN_FIELD_SIZE	sizeof(quint32)

///////////////////////////////////////////////////////////////////////////////
// class CNtRawData

class CNtRawData
{
	QByteArray* m_pStorage;
	int m_iDelta;

#include PRE_PACK
	struct HIVE_HEADER
	{
		quint32	Signature;
		quint32	Sequence1;
		quint32	Sequence2;
		quint8	TimeStamp[8];
		quint32	Major;
		quint32	Minor;
		quint32	Other1;
		quint32	Other2;
		quint32	RootCell;
		quint32	Length;
		quint32	Other3;
		quint8	Reserved[460];
		quint32	CheckSum;
	}PACKED;
#include POST_PACK

#include PRE_PACK
	struct BIN_HEADER
	{
		quint32	Signature;
		quint32	FileOffset;
		quint32	Size;
		quint32	Reserved[2];
		quint8	TimeStamp[8];
		quint32	Spare;
	}PACKED;
#include POST_PACK

	// Signature of hive header (0x66676572 = 'regf')
	const static quint32	ID_HIVE = 0x66676572;
	// Signature of bin header (0x6E696268 = 'hbin')
	const static quint32	ID_HBIN = 0x6E696268;

	qint32 CalcCheckSum();

	err_status_t FindFreeCell(qint32 dwSize, quint32& dwOffset);
	err_status_t AddBin(qint32 dwSize, quint32& dwOffset);

public:
	CNtRawData(QByteArray* pData, int iDelta) : m_pStorage(pData), m_iDelta(iDelta)
	{
		if(m_pStorage)
		{
			HIVE_HEADER*	pHdr = (HIVE_HEADER*) m_pStorage->data();

			LOG_MESSAGE(DBG_FATAL, "[CNtRawData::CNtRawData] File size = 0x%x, data size = 0x%x", m_pStorage->size(), pHdr->Length);
			WRITE_TRACE(DBG_FATAL, "OA00004.01:\t%x;\t%x", m_pStorage->size(), pHdr->Length);

			if(quint32(m_pStorage->size()) != pHdr->Length + HIVE_HDR_SIZE)
			{
				m_pStorage->resize(pHdr->Length + HIVE_HDR_SIZE);

				LOG_MESSAGE(DBG_FATAL, "[CNtRawData::CNtRawData] Storage was resized.");
				WRITE_TRACE(DBG_FATAL, "OA00004.02:");
			}

			m_pStorage->reserve(m_pStorage->size() + m_iDelta * 1024);
		}
	};
	virtual ~CNtRawData()
	{};

	err_status_t AllocCell(qint32 dwSize, quint32& dwOffset);
	err_status_t FreeCell(quint32 dwOffset);

	// Function returns true if this hive belongs to NT (and higher) registry
	bool IsNtHive() const
	{
		if(!m_pStorage)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtRawData::IsNtHive] m_pStorage is NULL.");
			WRITE_TRACE(DBG_FATAL, "OA00004.03:");

			return  false;
		}

		HIVE_HEADER*	pHdr = (HIVE_HEADER*) m_pStorage->data();

		return  (pHdr->Signature == ID_HIVE);
	}

	// Function returns major version of hive (or -1 in case of error)
	quint32 GetMajorVersion() const
	{
		if(!m_pStorage)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtRawData::GetMajorVersion] m_pStorage is NULL.");
			WRITE_TRACE(DBG_FATAL, "OA00004.04:");

			return  (quint32) -1;
		}

		HIVE_HEADER*	pHdr = (HIVE_HEADER*) m_pStorage->data();

		if(pHdr->Signature == ID_HIVE)
			return  pHdr->Major;

		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::GetMajorVersion] Unknown hive (0x%x).", pHdr->Signature);
		WRITE_TRACE(DBG_FATAL, "OA00004.05:\t%x", pHdr->Signature);

		return  (quint32) -1;
	}

	// Function returns minor version of hive (or -1 in case of error)
	quint32 GetMinorVersion() const
	{
		if(!m_pStorage)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtRawData::GetMinorVersion] pStorage is NULL.");
			WRITE_TRACE(DBG_FATAL, "OA00004.06:");

			return  (quint32) -1;
		}

		HIVE_HEADER*	pHdr = (HIVE_HEADER*) m_pStorage->data();

		if(pHdr->Signature == ID_HIVE)
			return  pHdr->Minor;

		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::GetMinorVersion] Unknown hive (0x%x).", pHdr->Signature);
		WRITE_TRACE(DBG_FATAL, "OA00004.07:\t%x", pHdr->Signature);

		return  (quint32) -1;
	}

	// Function returns relative offset of root key (or -1 in case of error)
	quint32 GetRootCellOffset() const
	{
		if(!m_pStorage)
		{
			LOG_MESSAGE(DBG_FATAL, "[CNtRawData::GetRootCellOffset] m_pStorage is NULL.");
			WRITE_TRACE(DBG_FATAL, "OA00004.08:");

			return  (quint32) -1;
		}

		HIVE_HEADER*	pHdr = (HIVE_HEADER*) m_pStorage->data();

		if(pHdr->Signature == ID_HIVE)
			return  pHdr->RootCell;

		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::GetRootCellOffset] Unknown hive (0x%x).", pHdr->Signature);
		WRITE_TRACE(DBG_FATAL, "OA00004.09:\t%x", pHdr->Signature);

		return  (quint32) -1;
	}

	// Function returns pointer to raw data of hive (or NULL in case of error)
	quint8* GetRawData()
	{
		if(m_pStorage)
			return  (quint8*) m_pStorage->data();

		LOG_MESSAGE(DBG_FATAL, "[CNtRawData::GetRawData] m_pStorage is NULL.");
		WRITE_TRACE(DBG_FATAL, "OA00004.10:");

		return  NULL;
	}
};

#endif // __NT__RAW__DATA__H__
