/////////////////////////////////////////////////////////////////////////////
///
/// @file QtHive.cpp
///
/// This module implements interface IHive via QFile.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#include <QFile>
#include <prlcommon/Logging/Logging.h>
#include "WinRegistryErrors.h"
#include "QtHive.h"

///////////////////////////////////////////////////////////////////////////////
// class CQtHive

CQtHive::CQtHive() : m_bDirty(false)
{
}


CQtHive::~CQtHive()
{
	Close();
}


/**
 * "Opens" specified hive, i.e. opens file contained hive, reads
 * its contents entirely to storage, closes file.
 *
 * @author igorl
 *
 * @param[in] sHiveName is name of file contained specified hive.
 *
 * @return PRL_ERR_SUCCESS if there is no any error or error code
 * otherwise.
 */
err_status_t CQtHive::open_(QFile& hive_)
{
	if (hive_.isOpen() && hive_.isReadable())
	{
		// More than int we could not process (see QByteArray)
		int iSize = static_cast<int>(hive_.size());

		// Read file whole
		hive_.seek(0);
		m_aData = hive_.readAll();

		if(iSize == m_aData.size())
		{
			m_sHiveName = hive_.fileName();
			return  PRL_ERR_SUCCESS;
		}
	}
	LOG_MESSAGE(DBG_FATAL, "[CQtHive::Open]"
				" Error during hive opening (%s).",
		qPrintable(hive_.fileName()));
	WRITE_TRACE(DBG_FATAL, "OA00006.01:\t%s", qPrintable(hive_.fileName()));

	return  WinRegistry::REG_IO_ERROR;
}

err_status_t CQtHive::Open(const QString& sHiveName)
{
	QFile h(sHiveName);
	h.open(QIODevice::ReadOnly);
	return open_(h);
}

err_status_t CQtHive::Open(FILE* hive_)
{
	QFile h;
	h.open(hive_, QIODevice::ReadOnly);
	return open_(h);
}

/**
 * Writes contents of storage back to specified hive, i.e. opens
 * file contained hive, truncates this file, writes contents of
 * storage to file, closes it.
 * P.S.
 * Writing is executed only if contents of storage was changed
 * (i.e. m_bDirty is true).
 *
 * @author igorl
 *
 * @return PRL_ERR_SUCCESS if there is no any error or error code
 * otherwise.
 */
err_status_t CQtHive::Write(void)
{
	if(!m_bDirty)
	{
		LOG_MESSAGE(DBG_FATAL, "[CQtHive::Write] Writing was skiped.");
		WRITE_TRACE(DBG_FATAL, "OA00006.02:");

		return  PRL_ERR_SUCCESS;
	}

	QFile  file(m_sHiveName);
	int    iWrote = 0;

	if(file.open(QIODevice::WriteOnly | QIODevice::Truncate))
	{
		file.seek(0);
		iWrote = static_cast<int>(file.write(m_aData));

		if(iWrote == m_aData.size())
			return  PRL_ERR_SUCCESS;
	}

	LOG_MESSAGE(DBG_FATAL, "[CQtHive::Write]"
		" Error during hive writing (its size %d, wrote %d).",
												m_aData.size(), iWrote);
	WRITE_TRACE(DBG_FATAL, "OA00006.03:\t%d;\t%d",
												m_aData.size(), iWrote);

	return  WinRegistry::REG_IO_ERROR;
}


/**
 * Closes hive called Write first.
 *
 * @author igorl
 *
 * @return PRL_ERR_SUCCESS if there is no any error or error code
 * otherwise.
 *
 */
err_status_t CQtHive::Close(void)
{
	return  Write();
}
