/////////////////////////////////////////////////////////////////////////////
///
/// @file QtHive.h
///
/// This module implements interface IHive via QFile.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#ifndef	__QT__HIVE__H__
#define	__QT__HIVE__H__

#include "IHive.h"
#include <QByteArray>

class QFile;
///////////////////////////////////////////////////////////////////////////////
// class CQtHive

class CQtHive : public IHive
{
	/// Name of file contained registry hive.
	QString			m_sHiveName;

	/// Storage contained hive itself.
	QByteArray		m_aData;

	/// This is the sign that hive data were changed. It is used for
	/// automatically hive saving in destructor.
	bool			m_bDirty;

public:
	CQtHive();
	virtual ~CQtHive();


	/**
	 * "Opens" specified hive, i.e. opens file contained hive, reads
	 * its contents entirely to storage, closes file.
	 *
	 * @author igorl
	 *
	 * @param[in] sHiveName is name of file contained specified hive.
	 *
	 * @return R_SUCCESS if there is no any error or error code
	 * otherwise.
	 */
	virtual err_status_t Open(const QString& sHiveName);
	virtual err_status_t Open(FILE* hive_);


	/**
	 * Writes contents of storage back to specified hive, i.e. opens
	 * file contained hive, truncates this file, writes contents of
	 * storage to file, closes it.
	 * P.S.
	 * Writing is executed only if contents of storage was changed
	 * (i.e. m_bDirty is true).
	 *
	 * @author igorl
	 *
	 * @return R_SUCCESS if there is no any error or error code
	 * otherwise.
	 */
	virtual err_status_t Write();


	/**
	 * Closes hive called Write first.
	 *
	 * @author igorl
	 *
	 * @return R_SUCCESS if there is no any error or error code
	 * otherwise.
	 *
	 */
	virtual err_status_t Close();


	/**
	 * Returns pointer to QByteArray contained hive.
	 *
	 * @author igorl
	 */
	virtual QByteArray* GetStorage()
	{
		return  &m_aData;
	}


	/**
	 * Deletes itself.
	 *
	 * @author igorl
	 */
	virtual void Free()
	{
		delete  this;
	}


	/**
	 * Returns true if contents of hive was changed or false
	 * otherwise.
	 *
	 * @author igorl
	 */
	bool IsHiveChanged() const
	{
		return  m_bDirty;
	}


	/**
	 * Changes dirty flag to specified.
	 *
	 * @author igorl
	 *
	 * @param[in] bDirty is a new value of dirty flag.
	 *
	 * @noreturn
	 */
	void SetDirtySign(bool bDirty = true)
	{
		m_bDirty = bDirty;
	}

private:
	err_status_t open_(QFile& hive_);
};

#endif // __QT__HIVE__H__
