/////////////////////////////////////////////////////////////////////////////
///
/// @file Registry.cpp
///
/// This module implements class for working with registry of any version
/// Windows (Win9X/ME/NT/2000/etc). The class implements basic functions for
/// manipulations of keys and values.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#include <QStack>
#include <stdio.h>
#include "QtHive.h"
#include "Registry.h"
#include <QStringList>
#include "NtBasicFunctions.h"
#include <prlcommon/Logging/Logging.h>


namespace {
	quint32 CalcChecksum(QByteArray* pStorage)
	{
		quint32 hash = 0;
		for (int i = 0; i < 508; i += sizeof(quint32))
			hash ^= *(quint32*)&pStorage->data()[i];

		return hash;
	}

	void CheckAndRecoverRegistry(QByteArray* pStorage)
	{
		quint32 seq1 = *(quint32*)&pStorage->data()[0x4];
		quint32 seq2 = *(quint32*)&pStorage->data()[0x8];
		quint32 checkSum = *(quint32*)&pStorage->data()[0x1fc];
		quint32 checkSum_calc = CalcChecksum(pStorage);

		if (seq1 == seq2 && checkSum == checkSum_calc)
			return;

		WRITE_TRACE(DBG_WARNING, "Registry file is corrupted: seq1=%u, seq2=%u, chksum=0x%x, chksum_calc=0x%x",
				seq1, seq2, checkSum, checkSum_calc);

		*(quint32*)&pStorage->data()[0x4] = seq2;
		checkSum_calc = CalcChecksum(pStorage);
		*(quint32*)&pStorage->data()[0x1fc] = checkSum_calc;

		WRITE_TRACE(DBG_WARNING, "Registry file after repair: seq1=%u, seq2=%u, chksum=0x%x, chksum_calc=0x%x",
				seq1, seq2, checkSum, checkSum_calc);
	}
}

///////////////////////////////////////////////////////////////////////////////
// class CRegistry

CRegistry::CRegistry(IHive* pHive, FILE* hive_, int iDelta)
: m_pHive(pHive)
, m_pFunctions(NULL)
{
	if(PRL_SUCCEEDED(m_pHive->Open(hive_)))
	{
		QByteArray*  pStorage = m_pHive->GetStorage();

		// Bugfix/workaround for #PSBM-25680
		// In some cases reg file can be in corrupted state after p2v
		// because of sequence numbers mismatch. Check seuence numbers
		// and hash and repair registry if needed.
		// NOTE: hive is not market as dirty to perform write only
		//       if data were actually modified.
		CheckAndRecoverRegistry(pStorage);

		switch(m_pHive->GetHiveVersion(reinterpret_cast<quint8*>(pStorage->data()), pStorage->size()))
		{
			case IHive::HV_WINNT:
				{
					m_pFunctions = new CNtBasicFunctions(pStorage, iDelta);
				}
				break;
			default:
				{
					LOG_MESSAGE(DBG_FATAL, "[CRegistry::CRegistry] Unsupported hive type.");
					WRITE_TRACE(DBG_FATAL, "OA00005.13:");
				}
				break;
		}
	}
	else
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::CRegistry] Some problem with opening Qt hive.");
		WRITE_TRACE(DBG_FATAL, "OA00005.07:");
	}
}


CRegistry::~CRegistry()
{
	if(m_pFunctions)
		m_pFunctions->Free();

	if(m_pHive)
		m_pHive->Free();
}

err_status_t CRegistry::OpenKey(const QString& sKeyName, quint32& dwKeyOffset, quint32 dwParentOffset)
{
	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::OpenKey] 'Executor' is epmty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.02:");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	return  m_pFunctions->OpenKey(sKeyName, dwKeyOffset, dwParentOffset);
}

err_status_t CRegistry::CreateKey(const QString& sKeyName, quint32& dwKeyOffset, quint32 dwParentOffset)
{
	QStringList		listNames = sKeyName.split("\\");
	err_status_t	Err;

	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::CreateKey] 'Executor' is epmty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.03:");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	for(int i = 0; i < listNames.size(); i++)
	{
		Err = m_pFunctions->OpenKey(listNames.at(i), dwKeyOffset, dwParentOffset);

		if(PRL_FAILED(Err))
		{
			if(Err == WinRegistry::REG_KEY_NOT_FOUND_ERROR)
			{
				Err = m_pFunctions->CreateKey(listNames.at(i), dwKeyOffset, dwParentOffset);

				if(PRL_FAILED(Err))
					return  Err;
			}
			else
			{
				return  Err;
			}
		}

		dwParentOffset = dwKeyOffset;
	}

	m_pHive->SetDirtySign();

	return  PRL_ERR_SUCCESS;
}


/**
 * Deletes a registry key with subkeys and values.
 *
 * @author mityak
 *
 * @param[in] qsKeyDeleteName is the name of the key to be deleted.
 *
 * @return PRL_ERR_SUCCESS if there is no any error or error code otherwise.
 */
err_status_t CRegistry::DeleteKey(const QString& qsKeyDeleteName)
{
	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::DeleteKey] 'Executor' is epmty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.04:");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	if ( qsKeyDeleteName.isEmpty() )
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::DeleteKey] Name of deleting key is empty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.08:");

		return WinRegistry::REG_KEY_NAME_EMPTY_ERROR;
	}

	err_status_t errorStatus;
	QStack<quint32> qStack;
	quint32 uiKeyOff = 0;

	// Check for existing delete name
	errorStatus = m_pFunctions->OpenKey( qsKeyDeleteName, uiKeyOff);

	if ( PRL_FAILED( errorStatus) )
		// Key does not exist
		return errorStatus;

	// Push opened deleting key
	qStack.push( uiKeyOff);
	qStack.push( 0);

	quint32 uiIndex = 0;
	QString qsKeyName = QString("");
	quint32 uiParentKeyOff;

	// Delete key
	while ( !qStack.isEmpty() )
	{
		// Get current deleting sub key
		uiIndex = qStack.pop();
		uiParentKeyOff = qStack.pop();

		// Delete previous subkey in curent subkey
		if ( uiIndex )
		{
			// Get deleting previous sub key
			errorStatus = m_pFunctions->EnumKey( uiParentKeyOff, --uiIndex, &qsKeyName, NULL);
			if ( PRL_FAILED( errorStatus) )
			{
				return errorStatus;
			}

			// Delete previous key
			errorStatus = m_pFunctions->DeleteKey( qsKeyName, uiParentKeyOff);
			if ( PRL_FAILED( errorStatus) )
			{
				return errorStatus;
			}
		}

		// Get next deleting sub key from current sub key
		errorStatus = m_pFunctions->EnumKey( uiParentKeyOff, uiIndex, &qsKeyName, NULL);

		if ( PRL_SUCCEEDED( errorStatus) )
		{
			// Next subkey is exist in current subkey

			// Open current subkey
			errorStatus = m_pFunctions->OpenKey( qsKeyName, uiKeyOff, uiParentKeyOff);
			if ( PRL_FAILED( errorStatus) )
			{
				return errorStatus;
			}

			// Push current subkey, and set him subkey as current subkey
			uiIndex++;
			qStack.push( uiParentKeyOff);
			qStack.push( uiIndex);

			// Current key is exist, go to in it
			qStack.push( uiKeyOff);
			qStack.push( 0);

			continue;
		}
		else
		{
			// Next subkey does not exist in current subkey
			if ( errorStatus != WinRegistry::REG_INDEX_TOO_LARGE )
			{
				return errorStatus;
			}

			// All keys in current key is passed, run through all values in current key
			quint32 uiIndVal = 0;
			quint32 uiType = RT_NONE;
			QString qsValueName = QString("");

			// Delete all values in current subkey
			while (1)
			{
				errorStatus = m_pFunctions->EnumValue( uiParentKeyOff, uiIndVal, &qsValueName, &uiType, NULL, NULL);

				// Check for end of enumeration
				if ( PRL_FAILED( errorStatus) )
					// End of enumeration
					break;

				errorStatus = m_pFunctions->DeleteValue( uiParentKeyOff, qsValueName);
				if ( PRL_FAILED( errorStatus) )
				{
					return errorStatus;
				}
			}
		}
	}

	// Delete master key
	QStringList qslSubKeys = qsKeyDeleteName.split("\\");
	QString qsSubKeyDel = qslSubKeys.last();

	qslSubKeys.removeLast();

	if ( qslSubKeys.isEmpty() )
	{
		uiParentKeyOff = -1;
	}
	else
	{
		qsKeyName = QString("");

		while ( !qslSubKeys.isEmpty() )
		{
			qsKeyName += qslSubKeys.first() + QString("\\");
			qslSubKeys.removeFirst();
		}

		qsKeyName.truncate( qsKeyName.length() - 1);

		errorStatus = m_pFunctions->OpenKey( qsKeyName, uiParentKeyOff);
		if ( PRL_FAILED( errorStatus) )
		{
			return errorStatus;
		}
	}

	errorStatus = m_pFunctions->DeleteKey( qsSubKeyDel, uiParentKeyOff);

	if ( PRL_FAILED( errorStatus) )
	{
		return errorStatus;
	}

	m_pHive->SetDirtySign();

	return PRL_ERR_SUCCESS;
}


/**
 * Copies the registry key qsOldName with subkeys and values to key qsNewName.
 * If a key with name qsNewName already exists CopyKey returns error.
 *
 * @author mityak
 *
 * @param[in] qsOldName is the name of the key to be copied.
 * @param[in] qsNewName is the name of target key.
 *
 * @return PRL_ERR_SUCCESS if there is no any error or error code otherwise.
 */
err_status_t CRegistry::CopyKey(const QString& qsOldName, const QString& qsNewName)
{
	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::CopyKey] 'Executor' is epmty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.09:");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	if ( qsNewName.isEmpty() )
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::CopyKey] New name is empty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.10:");

		return WinRegistry::REG_KEY_NAME_EMPTY_ERROR;
	}

	err_status_t errorStatus;
	QStack<quint32> qStack;
	QStack<quint32> qStackNew;
	quint32 uiKeyOff;

	// Check for existing old name
	errorStatus = m_pFunctions->OpenKey( qsOldName, uiKeyOff);

	if ( PRL_FAILED( errorStatus) )
		// Key does not exist
		return errorStatus;

	// Check for equal names
	if ( !qsOldName.compare(qsNewName, Qt::CaseInsensitive) )
		// Names are equal, nothing to do
		return WinRegistry::REG_KEY_ALREADY_EXISTS_ERROR;

	// Push opened old key and push current checking index of subkey in current key
	qStack.push( uiKeyOff);
	qStack.push( 0);

	// Check for exist new name already
	errorStatus = m_pFunctions->OpenKey( qsNewName, uiKeyOff);

	if ( PRL_SUCCEEDED( errorStatus) )
		// New key is existed already
		return WinRegistry::REG_KEY_ALREADY_EXISTS_ERROR;

	// Create new key
	QStringList	listNames = qsNewName.split("\\");
	quint32 uiSubKeyOff = 0;
	quint32 uiParentKeyOff = -1;

	for (int i = 0; i < listNames.size(); i++)
	{
		errorStatus = m_pFunctions->OpenKey(listNames.at(i), uiSubKeyOff, uiParentKeyOff);
		if( PRL_FAILED(errorStatus) )
		{
			if( errorStatus == WinRegistry::REG_KEY_NOT_FOUND_ERROR )
			{
				errorStatus = m_pFunctions->CreateKey(listNames.at(i), uiSubKeyOff, uiParentKeyOff);
				if( PRL_FAILED(errorStatus) )
					return  errorStatus;
			}
			else
				return  errorStatus;
		}

		uiParentKeyOff = uiSubKeyOff;
	}

	if ( PRL_FAILED( errorStatus) )
		return errorStatus;

	// Store id of new key
	qStackNew.push( uiParentKeyOff);

	quint32 uiIndex = 0;
	QString qsKeyName = QString("");

	// Create new name
	while ( !qStack.isEmpty() )
	{
		// Get current old sub key
		uiIndex = qStack.pop();
		uiParentKeyOff = qStack.pop();

		// Get next old sub key from current old sub key
		errorStatus = m_pFunctions->EnumKey( uiParentKeyOff, uiIndex, &qsKeyName, NULL);

		if ( PRL_SUCCEEDED( errorStatus) )
		{
			// Next subkey is exist in current subkey

			// Open current subkey
			errorStatus = m_pFunctions->OpenKey( qsKeyName, uiKeyOff, uiParentKeyOff);

			if ( PRL_FAILED( errorStatus) )
			{
				return errorStatus;
			}

			// Push current subkey, and set next subkey as current subkey
			uiIndex++;
			qStack.push( uiParentKeyOff);
			qStack.push( uiIndex);

			// Current key is exist, go to in it
			qStack.push( uiKeyOff);
			qStack.push( 0);

			// Create copy of this next old key in new subkey
			uiParentKeyOff = qStackNew.pop();

			// Create copy of next old key in new key
			errorStatus = m_pFunctions->CreateKey( qsKeyName, uiKeyOff, uiParentKeyOff);

			if ( PRL_FAILED( errorStatus) )
			{
				return errorStatus;
			}

			qStackNew.push( uiParentKeyOff);
			qStackNew.push( uiKeyOff);

			continue;
		}
		else
		{
			// Next subkey does not exist in current subkey
			if ( errorStatus != WinRegistry::REG_INDEX_TOO_LARGE )
			{
				return errorStatus;
			}

			// All keys in current key is passed, run through all values in current key
			quint32 uiIndVal = 0;
			quint32 uiType = RT_NONE;
			QString qsValueName = QString("");
			QByteArray qsValueData;
			quint32 uiSize, uiRealSize;

			uiKeyOff = qStackNew.pop();

			// Copy all values in current subkey
			while (1)
			{
				uiType = 0;
				errorStatus = m_pFunctions->EnumValue( uiParentKeyOff, uiIndVal, &qsValueName, &uiType, NULL, NULL);

				// Check for end of enumeration
				if ( PRL_FAILED( errorStatus) )
					// End of enumeration
					break;

				// Coping value

				qsValueData.resize( 1);
				errorStatus =
					m_pFunctions->GetValueData( uiParentKeyOff, qsValueName, uiType,
														(quint8 *)qsValueData.data(), 1, uiRealSize);
				if ( PRL_FAILED( errorStatus) )
				{
					if ( errorStatus != WinRegistry::REG_BUFFER_TOO_SMALL_ERROR )
					{
						return errorStatus;
					}
				}

				qsValueData.resize( uiRealSize);
				errorStatus =
					m_pFunctions->GetValueData( uiParentKeyOff, qsValueName, uiType,
														(quint8 *)qsValueData.data(), uiRealSize, uiSize);

				errorStatus = m_pFunctions->CreateValue( uiKeyOff, qsValueName, uiType);
				if ( PRL_FAILED( errorStatus) )
				{
					return errorStatus;
				}

				errorStatus =
					m_pFunctions->SetValueData( uiKeyOff, qsValueName, uiType,
														(quint8 *)qsValueData.data(), uiRealSize);
				if ( PRL_FAILED( errorStatus) )
				{
					return errorStatus;
				}

				uiIndVal++;
			}
		}
	}

	m_pHive->SetDirtySign();

	return PRL_ERR_SUCCESS;
}


/**
 * Renames the registry key qsOldName to qsNewName.
 *
 * @author mityak
 *
 * @param[in] qsOldName is the name of the key to be renamed.
 * @param[in] qsNewName is the new name of key.
 *
 * @return PRL_ERR_SUCCESS if there is no any error or error code otherwise.
 */
err_status_t CRegistry::RenameKey(const QString& qsOldName, const QString& qsNewName)
{
	err_status_t errorStatus;

	errorStatus = CopyKey(qsOldName, qsNewName);
	if ( PRL_FAILED(errorStatus) )
		return errorStatus;

	errorStatus = DeleteKey(qsOldName);
	if ( PRL_FAILED(errorStatus) )
		return errorStatus;

	return PRL_ERR_SUCCESS;
}


err_status_t CRegistry::QueryInfoKey(quint32 dwOffset, QString* psClassName, quint32* pcSubKeys, quint16* pMaxSubKeyLen, quint16* pMaxClassLen, quint32* pcValues, quint16* pMaxValueLen, quint32* pMaxValueDataLen)
{
	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::QueryInfoKey] 'Executor' is epmty.");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	return  m_pFunctions->QueryInfoKey(dwOffset, psClassName, pcSubKeys, pMaxSubKeyLen, pMaxClassLen, pcValues, pMaxValueLen, pMaxValueDataLen);
}

err_status_t CRegistry::EnumKey(quint32 dwOffset, quint32 dwIndex, QString* psName, QString* psClass)
{
	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::GetValueData] 'Executor' is epmty.");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	return  m_pFunctions->EnumKey(dwOffset, dwIndex, psName, psClass);
}

err_status_t CRegistry::GetValueData(quint32 dwNKoffset, const QString& sValueName, quint32 dwValueType, quint8* pBuf, quint32 dwBufSize, quint32& dwRealDataSize, quint32* pdwValueType)
{
	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::GetValueData] 'Executor' is epmty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.05:");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	return  m_pFunctions->GetValueData(dwNKoffset, sValueName, dwValueType, pBuf, dwBufSize, dwRealDataSize, pdwValueType);
}

err_status_t CRegistry::SetValueData(quint32 dwNKoffset, const QString& sValueName, quint32 dwValueType, quint8* pBuf, quint32 dwBufSize)
{
	err_status_t	Err;

	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::SetValueData] 'Executor' is epmty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.06:");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	Err = m_pFunctions->SetValueData(dwNKoffset, sValueName, dwValueType, pBuf, dwBufSize);

	if(PRL_FAILED(Err))
	{
		if(Err == WinRegistry::REG_VALUE_NOT_EXIST_ERROR)
		{
			Err = m_pFunctions->CreateValue(dwNKoffset, sValueName, dwValueType);

			if(PRL_SUCCEEDED(Err))
			{
				Err = m_pFunctions->SetValueData(dwNKoffset, sValueName, dwValueType, pBuf, dwBufSize);
			}
		}
	}

	if(PRL_SUCCEEDED(Err))
		m_pHive->SetDirtySign();

	return  Err;
}

err_status_t CRegistry::EnumValue(quint32 dwOffset, quint32 dwIndex, QString* psValue, quint32* pdwType, quint8* pData, quint32* pdwDataSize)
{
	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::GetValueData] 'Executor' is epmty.");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	return  m_pFunctions->EnumValue(dwOffset, dwIndex, psValue, pdwType, pData, pdwDataSize);
}


/**
 * Renames the value qsOldName to qsNewName.
 *
 * @author mityak
 *
 * @param[in] qsOldName is the name of the value to be renamed.
 * @param[in] qsNewName is the new name of value.
 *
 * @return PRL_ERR_SUCCESS if there is no any error or error code otherwise.
 */
err_status_t CRegistry::RenameValue(quint32 dwNKoffset, const QString& sOldValueName, const QString& sNewValueName)
{
	if(!m_pFunctions)
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::RenameValue] 'Executor' is epmty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.11:");

		return  WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	if ( sNewValueName.isEmpty() )
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::RenameValue] New name is empty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.12:");

		return WinRegistry::REG_VALUE_NOT_EXIST_ERROR;
	}

	quint32 dwValueType = 0;
	quint32 dwBufSize = 0;
	quint32 dwRealDataSize = 0;
	quint8  *pBuf = NULL;

	err_status_t err;
	quint8 c;

	// Find size of value data
	// FIXME:
	// Using of dwValueType is half hack solution! Please, change this place when someone
	// implements correct way to determine type of value.
	err = m_pFunctions->GetValueData(dwNKoffset, sOldValueName, 0, &c, 1, dwRealDataSize, &dwValueType);
	if ( PRL_FAILED(err) && (err != WinRegistry::REG_BUFFER_TOO_SMALL_ERROR) )
		return err;

	dwBufSize = dwRealDataSize;

	err = m_pFunctions->CreateValue(dwNKoffset, sNewValueName, dwValueType);
	if ( PRL_FAILED(err) )
		return err;

	pBuf = new quint8[dwBufSize];

	err = m_pFunctions->GetValueData(dwNKoffset, sOldValueName, 0, pBuf, dwBufSize, dwRealDataSize);
	if ( PRL_FAILED(err) )
	{
		delete pBuf;
		return err;
	}

	err = m_pFunctions->SetValueData(dwNKoffset, sNewValueName, dwValueType, pBuf, dwBufSize);

	delete pBuf;

	if ( PRL_FAILED(err) )
		return err;

	err = m_pFunctions->DeleteValue(dwNKoffset, sOldValueName);
	if ( PRL_FAILED(err) )
		return err;

	m_pHive->SetDirtySign();

	return  PRL_ERR_SUCCESS;
}


/**
 * Removes a named value from the specified registry key.
 *
 * @author igorl
 *
 * @param[in] dwNKoffset is a "handle" to an open registry key.
 * @param[in] sValueName is the registry value to be removed.
 *
 * @return PRL_ERR_SUCCESS if there is no any error or error code otherwise.
 */
err_status_t CRegistry::DeleteValue(quint32 dwNKoffset, const QString& sValueName)
{
	if ( !m_pFunctions )
	{
		LOG_MESSAGE(DBG_FATAL, "[CRegistry::DeleteValue] 'Executor' is epmty.");
		WRITE_TRACE(DBG_FATAL, "OA00005.15:");

		return WinRegistry::REG_EMPTY_EXECUTOR_ERROR;
	}

	err_status_t error = m_pFunctions->DeleteValue( dwNKoffset, sValueName );
	if ( PRL_SUCCEEDED(error) )
		m_pHive->SetDirtySign();

	return error;
}

///////////////////////////////////////////////////////////////////////////////
// struct CRegistryQtFile

CRegistryQtFile::CRegistryQtFile(FILE* hive_, int iDelta):
	CRegistry(new CQtHive(), hive_, iDelta), m_guard(hive_)
{
}

///////////////////////////////////////////////////////////////////////////////
// struct CRegistryQtFile::Sweeper

void CRegistryQtFile::Sweeper::cleanup(FILE* victim_)
{
	fclose(victim_);
}

