/////////////////////////////////////////////////////////////////////////////
///
/// @file Registry.h
///
/// This module implements class for working with registry of any version
/// Windows (Win9X/ME/NT/2000/etc). The class implements basic functions for
/// manipulations of keys and values.
///
/// @author igorl
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#ifndef	__REGISTRY__H__
#define	__REGISTRY__H__

#include "IHive.h"
#include "RegTypes.h"
#include <QScopedPointer>
#include "IBasicFunctions.h"
#include <boost/noncopyable.hpp>

///////////////////////////////////////////////////////////////////////////////
// class CRegistry

class CRegistry
{
	IHive* m_pHive;
	IBasicFunctions	*m_pFunctions;

public:
	CRegistry(IHive* pHive, FILE* hive_, int iDelta = 512);
	virtual ~CRegistry();

	err_status_t OpenKey(const QString& sKeyName, quint32& dwKeyOffset, quint32 dwParentOffset = (quint32) -1);
	err_status_t CreateKey(const QString& sKeyName, quint32& dwKeyOffset, quint32 dwParentOffset = (quint32) -1);


	/**
	 * Deletes a registry key with subkeys and values.
	 *
	 * @author mityak
	 *
	 * @param[in] qsKeyDeleteName is the name of the key to be deleted.
	 *
	 * @return R_SUCCESS if there is no any error or error code otherwise.
	 */
	err_status_t DeleteKey(const QString& sKeyName);


	/**
	 * Copies the registry key qsOldName with subkeys and values to key qsNewName.
	 * If a key with name qsNewName already exists CopyKey returns error.
	 *
	 * @author mityak
	 *
	 * @param[in] qsOldName is the name of the key to be copied.
	 * @param[in] qsNewName is the name of target key.
	 *
	 * @return R_SUCCESS if there is no any error or error code otherwise.
	 */
	err_status_t CopyKey(const QString& qsOldName, const QString& qsNewName);


	/**
	 * Renames the registry key qsOldName to qsNewName.
	 *
	 * @author mityak
	 *
	 * @param[in] qsOldName is the name of the key to be renamed.
	 * @param[in] qsNewName is the new name of key.
	 *
	 * @return R_SUCCESS if there is no any error or error code otherwise.
	 */
	err_status_t RenameKey(const QString& qsOldName, const QString& qsNewName);


	err_status_t QueryInfoKey(quint32 dwOffset, QString* psClassName, quint32* pcSubKeys, quint16* pMaxSubKeyLen, quint16* pMaxClassLen, quint32* pcValues, quint16* pMaxValueLen, quint32* pMaxValueDataLen);
	err_status_t EnumKey(quint32 dwOffset, quint32 dwIndex, QString* psName, QString* psClass);

	err_status_t GetValueData(quint32 dwNKoffset, const QString& sValueName, quint32 dwValueType, quint8* pBuf, quint32 dwBufSize, quint32& dwRealDataSize, quint32* pdwValueType = NULL);
	err_status_t SetValueData(quint32 dwNKoffset, const QString& sValueName, quint32 dwValueType, quint8* pBuf, quint32 dwBufSize);
	err_status_t EnumValue(quint32 dwOffset, quint32 dwIndex, QString* psValue, quint32* pdwType, quint8* pData, quint32* pdwDataSize);


	/**
	 * Renames the value qsOldName to qsNewName.
	 *
	 * @author mityak
	 *
	 * @param[in] qsOldName is the name of the value to be renamed.
	 * @param[in] qsNewName is the new name of value.
	 *
	 * @return R_SUCCESS if there is no any error or error code otherwise.
	 */
	err_status_t RenameValue(quint32 dwNKoffset, const QString& sOldValueName, const QString& sNewValueName);


	/**
	 * Removes a named value from the specified registry key.
	 *
	 * @author igorl
	 *
	 * @param[in] dwNKoffset is a "handle" to an open registry key.
	 * @param[in] sValueName is the registry value to be removed.
	 *
	 * @return R_SUCCESS if there is no any error or error code otherwise.
	 */
	err_status_t DeleteValue(quint32 dwNKoffset, const QString& sValueName);
};

///////////////////////////////////////////////////////////////////////////////
// struct CRegistryQtFile

struct CRegistryQtFile: CRegistry, boost::noncopyable
{
	CRegistryQtFile(FILE* hive_, int iDelta = 512);

private:
	struct Sweeper
	{
		static void cleanup(FILE* victim_);
	};

	QScopedPointer<FILE, Sweeper> m_guard;
};

#endif // __REGISTRY__H__

