/////////////////////////////////////////////////////////////////////////////
///
/// @file Windows.cpp
///
/// VM authentication for Windows OSes.
///
/// @author shrike
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2020 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#include <openssl/md4.h>
#include <openssl/md5.h>
#include <openssl/rc4.h>
#include <openssl/aes.h>
#include <openssl/err.h>
#include <boost/program_options.hpp>
#include <prlcommon/Logging/Logging.h>

#include "Windows.h"
#include "Facade.h"

namespace Windows
{

const QString User::getName() const
{
	return QString(reinterpret_cast<const QChar *>
		(m_info->data + m_info->username_shift),
		m_info->username_size >> 1);
}

quint8 const (& User::getPasswordHash() const)[0x10]
{
	const PSWD_CRYPTED_DATA* d = reinterpret_cast<const PSWD_CRYPTED_DATA* >
		(getPasswordData());

	return d->Hash;
}

const uint8_t* User::getPasswordData() const
{
	return m_info->data + m_info->ntpw_shift;
}

QByteArray AES::operator()(const char* key_) const
{
	EVP_CIPHER_CTX* x = EVP_CIPHER_CTX_new();
	if (NULL != x)
	{
		Prl::Expected<QByteArray, int> y = decrypt(key_, x);
		if (y.isSucceed())
		{
			EVP_CIPHER_CTX_free(x);
			return y.value();
		}
	}
	for (unsigned long e; (e = ERR_get_error());)
	{
		char *err = ERR_error_string(e, NULL);
		WRITE_TRACE(DBG_FATAL, "%s\n", err);
	}
	if (x)
		EVP_CIPHER_CTX_free(x);

	return QByteArray();
}

Prl::Expected<QByteArray, int>
	AES::decrypt(const char* key_, EVP_CIPHER_CTX* ctx_) const
{
	if (!EVP_DecryptInit_ex(ctx_, EVP_aes_128_cbc(), NULL,
		reinterpret_cast<const uint8_t* >(key_), m_salt))
		return -1;

	int len = 0;
	EVP_CIPHER_CTX_set_key_length(ctx_, EVP_MAX_KEY_LENGTH);
	QByteArray output(m_length, 0);
	uint8_t* plaintext = reinterpret_cast<uint8_t* >(output.data());
	if (EVP_DecryptUpdate(ctx_, plaintext, &len, m_data, m_length) != 1)
		return -1;

	int total = len;
	if (EVP_DecryptFinal_ex(ctx_, plaintext + len, &len) < 0)
		return -1;

	total += len;
	output.resize(total);
	return output;
}

///////////////////////////////////////////////////////////////////////////////
// struct Magic

const char Magic::s_ANUM[] = "0123456789012345678901234567890123456789";
const char Magic::s_AQWERTY[] = "!@#$%^&*()qwertyUIOPAzxcvbnmQQQQQQQQQQQQ)(*@&%";
const char Magic::s_ANTPASSWORD[] = "NTPASSWORD";

const_DES_cblock* Magic::generate(quint8 const (&src_)[7], DES_cblock& dst_)
{
	dst_[0] = src_[0];
	dst_[1] = (((src_[0] & 0x01) << 6) | (src_[1] >> 2)) << 1;
	dst_[2] = (((src_[1] & 0x03) << 5) | (src_[2] >> 3)) << 1;
	dst_[3] = (((src_[2] & 0x07) << 4) | (src_[3] >> 4)) << 1;	dst_[4] = (((src_[3] & 0x0F) << 3) | (src_[4] >> 5)) << 1;
	dst_[4] = (((src_[3] & 0x0F) << 3) | (src_[4] >> 5)) << 1;
	dst_[5] = (((src_[4] & 0x1F) << 2) | (src_[5] >> 6)) << 1;
	dst_[6] = (((src_[5] & 0x3F) << 1) | (src_[6] >> 7)) << 1;
	dst_[7] = (src_[6] & 0x7F) << 1;
	DES_set_odd_parity(&dst_);

	return &dst_;
}

QByteArray Magic::hash(quint8 const (&bootkey_)[0x30], const QByteArray& syskey_)
{
	MD5_CTX md5c;
	RC4_KEY rc4k;
	quint8 md5hash[0x10];

	MD5_Init(&md5c);
	MD5_Update(&md5c, bootkey_, 0x10);
	MD5_Update(&md5c, s_AQWERTY, sizeof(s_AQWERTY));
	MD5_Update(&md5c, syskey_.data(), syskey_.size());
	MD5_Update(&md5c, s_ANUM, sizeof(s_ANUM));
	MD5_Final(md5hash, &md5c);

	QByteArray output(0x20, 0);
	RC4_set_key(&rc4k, sizeof(md5hash), md5hash);
	RC4(&rc4k, output.size(), &bootkey_[0x10],
		reinterpret_cast<quint8* >(output.data()));

	return output;
}

QByteArray Magic::decode(const QByteArray& key_, const Account& for_)
{
	const SAM_HASH* h = reinterpret_cast<const SAM_HASH* >
			(for_.getPasswordData());

	switch (h->revision)
	{
	case 1:
	{
		MD5_CTX md5c;
		RC4_KEY rc4k;
		quint8 md5hash[0x10];

		MD5_Init(&md5c);
		quint32 r = for_.getRid();
		MD5_Update(&md5c, key_.data(), 0x10);
		MD5_Update(&md5c, reinterpret_cast<const quint8* >(&r), sizeof(r));
		MD5_Update(&md5c, s_ANTPASSWORD, sizeof(s_ANTPASSWORD));
		MD5_Final(md5hash, &md5c);

		QByteArray output(sizeof(for_.getPasswordHash()), 0);
		RC4_set_key(&rc4k, sizeof(md5hash), md5hash);
		RC4(&rc4k, output.size(), &for_.getPasswordHash()[0],
				reinterpret_cast<quint8* >(output.data()));
		return output;
	}
	case 2:
	{
		const SAM_HASH_AES *a = reinterpret_cast<const SAM_HASH_AES *>
				(for_.getPasswordData());
		Windows::AES aes(a->data, for_.getPasswordSize() -
				offsetof(SAM_HASH_AES, data), a->salt);
		return aes(key_.data());
	}
	default:
		WRITE_TRACE(DBG_FATAL, "Unknown SAM hash revision %d",
				h->revision);
		return QByteArray();
	}
}

QByteArray Magic::decrypt(const QByteArray& hash_, const Account& for_)
{
	DES_key_schedule des_key1 = DES_key_schedule(), des_key2 = DES_key_schedule();
	DES_cblock des_cb1 = {}, des_cb2 = {};

	quint32 r = for_.getRid();
	quint8 x[7] = {quint8(r), quint8(r >> 8), quint8(r >> 16), quint8(r >> 24)};
	x[4] = x[0];
	x[5] = x[1];
	x[6] = x[2];
	DES_set_key(generate(x, des_cb1), &des_key1);

	quint8 y[7] = {quint8(r >> 24), quint8(r), quint8(r >> 8), quint8(r >> 16)};
	y[4] = y[0];
	y[5] = y[1];
	y[6] = y[2];
	DES_set_key(generate(y, des_cb2), &des_key2);

	QByteArray output(hash_.size(), 0);
	DES_cblock (&o)[2] = *reinterpret_cast<DES_cblock (*)[2]>(output.data());
	DES_cblock (&i)[2] = *reinterpret_cast<DES_cblock (*)[2]>
		(const_cast<char* >(hash_.data()));
	DES_ecb_encrypt(&i[0], &o[0], &des_key1, DES_DECRYPT);
	DES_ecb_encrypt(&i[1], &o[1], &des_key2, DES_DECRYPT);

	return output;
}

QByteArray Magic::hash(const QString& password_)
{
	MD4_CTX x;
	MD4_Init(&x);
	MD4_Update(&x, password_.utf16(), password_.size() * sizeof(ushort));
	QByteArray output(0x10, 0);
	MD4_Final(reinterpret_cast<quint8* >(output.data()), &x);

	return output;
}

Prl::Expected<QByteArray, int> Magic::getSamKey(QByteArray& domAccF_,
		QByteArray& key_, int pwLen_)
{
	char *p = domAccF_.data();
	uint16_t rev = *p;

	if (rev == 2)
	{
		if (sizeof(PSWD_CRYPTED_DATA) == pwLen_)
			return hash(*reinterpret_cast<quint8 const (*)[0x30]>(p + 0x70), key_);

		WRITE_TRACE(DBG_FATAL, "password length mismatch (%d bytes)",
				pwLen_);
	}
	else if (rev == 3)
	{
		SAM_KEY_DATA_AES *a = reinterpret_cast<SAM_KEY_DATA_AES* >(p + 0x68);

		if (a->revision == 2)
		{
			Windows::AES aes(a->data, a->datalen, a->salt);
			return aes(key_.data());
		}
		WRITE_TRACE(DBG_FATAL, "Unknown AES key rvision %d",
				a->revision);
	}
	else
		WRITE_TRACE(DBG_FATAL, "Unknown Account/F revision %d", rev);

	return -1;
}

///////////////////////////////////////////////////////////////////////////////
// struct SAM definition

Prl::Expected<QByteArray, int> SAM::getF()
{
	quint32 a;
	if (PRL_FAILED(m_dataSource->OpenKey("SAM\\Domains\\Account", a)))
		return -1;

	quint32 r = 0x1000;
	QByteArray output(r, 0);
	if (PRL_FAILED(m_dataSource->GetValueData(a, "F", 0,
		reinterpret_cast<quint8* >(output.data()), r, r)))
		return -1;

	output.resize(r);
	return output;
}

Prl::Expected<Account, int> SAM::getV(const std::string& username_)
{
	quint32 a;
	if (PRL_FAILED(m_dataSource->OpenKey("SAM\\Domains\\Account\\Users", a)))
		return -1;

	QString k;
	for (quint32 i = 0; PRL_SUCCEEDED(m_dataSource->EnumKey(a, i, &k, NULL)); ++i)
	{
		if ("Names" == k)
			continue;

		quint32 h;
		if (PRL_FAILED(m_dataSource->OpenKey(k, h, a)))
			continue;

		quint32 r = 0x1000;
		QByteArray v(r, 0);
		if (PRL_FAILED(m_dataSource->GetValueData(h, "V", 0,
			reinterpret_cast<quint8* >(v.data()), r, r)))
			return -1;

		v.resize(r);
		User u(v);
		QString n = u.getName();
		WRITE_TRACE(DBG_INFO, "user (%d/%d) %s rid=%s",
			n.size(), username_.size(), qPrintable(n), qPrintable(k));
		if (!(username_.size() == static_cast<std::size_t>(n.size()) &&
			n.compare(username_.data(), Qt::CaseInsensitive) == 0))
			continue;
                if (1 != sscanf(qPrintable(k), "%X", &r))
		{
			WRITE_TRACE(DBG_FATAL, "failed to get RID for %s", qPrintable(k));
			return -1;
		}
		return Account(r, u);
	}
	WRITE_TRACE(DBG_FATAL, "Cannot find user's %s account record", username_.data());
	return -5;
}

///////////////////////////////////////////////////////////////////////////////
// struct Engine definition

const char* Engine::s_LSAS[] = {"JD", "Skew1", "GBG", "Data"};
const int Engine::s_PERMUTATION[0x10] =
	{0xb, 0x6, 0x7, 0x1, 0x8, 0xa, 0xe, 0x0, 0x3, 0x5, 0x2, 0xf, 0xd, 0x9, 0xc, 0x4};

int Engine::open(const char* path_, QScopedPointer<CRegistry>& dst_) const
{
	std::string x = m_root + path_;
	char* p = guestfs_case_sensitive_path(m_handle, x.data());
	if (NULL == p)
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_case_sensitive_path: %s", guestfs_last_error(m_handle));
		return -1;
	}
	Prl::Expected<FILE*, int> s = Facade(m_handle).download(p);
	free(p);
	if (s.isFailed())
		return s.error();

	dst_.reset(new CRegistryQtFile(s.value()));
	return 0;
}

Prl::Expected<SAM, int> Engine::getSAM() const
{
	QScopedPointer<CRegistry> h;
	int e = open("/system32/config/sam", h);
	if (0 != e)
		return e;

	return SAM(h.take());
}

Prl::Expected<QByteArray, int> Engine::getSyskey() const
{
	QScopedPointer<CRegistry> h;
	int e = open("/system32/config/system", h);
	if (0 != e)
		return e;

	quint32 k, z;
	if (PRL_FAILED(h->OpenKey("ControlSet001\\Control\\Lsa", k)))
		return -1;

	QByteArray t(0x10, 0);
	for (int i = boost::size(s_LSAS); i-- > 0;)
	{
		if (PRL_FAILED(h->OpenKey(s_LSAS[i], z, k)))
			return -1;

		QString u;
		if (PRL_FAILED(h->QueryInfoKey(z, &u, NULL, NULL, NULL, NULL, NULL, NULL)))
			return -1;

		QByteArray a = u.toUtf8();
		a[8] = 0;
		quint32 (&x)[4] = *reinterpret_cast<quint32(*)[4]>(t.data());
		if (1 != sscanf(a.constData(), "%x", &x[i]))
		{
			WRITE_TRACE(DBG_FATAL, "sscanf %s [%d]\n", a.constData(), errno);
			return -1;
		}
	}
	QByteArray output;
	for (int i = 0; i < t.size(); ++i)
	{
		output.push_back(t.at(s_PERMUTATION[i]));
	}

	return output;
}

} // namespace Windows
