/////////////////////////////////////////////////////////////////////////////
///
/// @file Windows.h
///
/// This module implements classes for Windows VM authentication.
///
/// @author shrike
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2020 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#ifndef	__WINDOWS__H__
#define	__WINDOWS__H__

#include <openssl/des.h>
#include <openssl/evp.h>
#include <guestfs.h>
#include <QByteArray>
#include <QSharedPointer>
#include <prlcommon/PrlCommonUtilsBase/SysError.h>

#include "Registry.h"
#include "wincrypt.h"

namespace Windows
{
struct Account;
///////////////////////////////////////////////////////////////////////////////
// struct SAM declaration

struct SAM
{
	explicit SAM(CRegistry* dataSource_): m_dataSource(dataSource_)
	{
	}

	Prl::Expected<QByteArray, int> getF();
	Prl::Expected<Account, int>
		getV(const std::string& username_);

private:
	QSharedPointer<CRegistry> m_dataSource;
};

///////////////////////////////////////////////////////////////////////////////
// struct Engine declaration

struct Engine
{
	Engine(guestfs_h* handle_, const std::string& root_):
		m_handle(handle_), m_root(root_)
	{
	}

	Prl::Expected<SAM, int> getSAM() const;
	Prl::Expected<QByteArray, int> getSyskey() const;

private:
	int open(const char* path_, QScopedPointer<CRegistry>& dst_) const;

	static const char* s_LSAS[];
	static const int s_PERMUTATION[0x10];

	guestfs_h* m_handle;
	const std::string m_root;
};

///////////////////////////////////////////////////////////////////////////////
// struct User

struct User
{
	explicit User(const QByteArray& store_): m_store(store_),
		m_info(reinterpret_cast<const USER_V* >(m_store.data()))
	{
	}

	const QString getName() const;
	quint32 getPasswordSize() const
	{
		return m_info->ntpw_size;
	}
	const uint8_t* getPasswordData() const;
	quint8 const (& getPasswordHash() const)[0x10];

private:
	const QByteArray m_store;
	const USER_V* m_info;
};

///////////////////////////////////////////////////////////////////////////////
// struct Account

struct Account: User
{
	Account(quint32 rid_, const User& user_): User(user_), m_rid(rid_)
	{
	}

	quint32 getRid() const
	{
		return m_rid;
	}

private:
	const quint32 m_rid;
};

///////////////////////////////////////////////////////////////////////////////
// struct AES

struct AES
{
	AES(const uint8_t* data_, int length_, const uint8_t* salt_):
		m_data(data_), m_length(length_), m_salt(salt_)
	{
	}

	QByteArray operator()(const char* key_) const;

private:
	Prl::Expected<QByteArray, int> decrypt
		(const char* key_, EVP_CIPHER_CTX *ctx_) const;

	const uint8_t* m_data;
	uint32_t m_length;
	const uint8_t* m_salt;
};

///////////////////////////////////////////////////////////////////////////////
// struct Magic

struct Magic
{
	QByteArray hash(const QString& password_);
	QByteArray hash(quint8 const (&bootkey_)[0x30], const QByteArray& syskey_);
	QByteArray decode(const QByteArray& bootkey_, const Account& for_);
	QByteArray decrypt(const QByteArray& hash_, const Account& for_);
	Prl::Expected<QByteArray, int> getSamKey(QByteArray& domAccF_,
			QByteArray& syskey_, int pwLen_);
private:
	static const_DES_cblock* generate(quint8 const (&src_)[7], DES_cblock& dst_);

	static const char s_ANUM[];
	static const char s_AQWERTY[];
	static const char s_ANTPASSWORD[];
};

} // namespace Windows

#endif // __WINDOWS__H__
