/////////////////////////////////////////////////////////////////////////////
///
/// @file main.cpp
///
/// VM authentication command line.
///
/// @author shrike
///
/// Copyright (c) 2005-2017, Parallels International GmbH
/// Copyright (c) 2017-2019 Virtuozzo International GmbH, All rights reserved.
///
/// This file is part of Virtuozzo Core. Virtuozzo Core is free
/// software; you can redistribute it and/or modify it under the terms
/// of the GNU General Public License as published by the Free Software
/// Foundation; either version 2 of the License, or (at your option) any
/// later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
/// 02110-1301, USA.
///
/// Our contact details: Virtuozzo International GmbH, Vordergasse 59, 8200
/// Schaffhausen, Switzerland.
///
///////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <shadow.h>
#include <guestfs.h>

#include <boost/foreach.hpp>
#include <boost/program_options.hpp>
#include <QByteArray>
#include <prlcommon/Logging/Logging.h>
#include <prlcommon/PrlCommonUtilsBase/SysError.h>

#include "Facade.h"

namespace po = boost::program_options;

namespace
{
void printHex(const char *name_, const QByteArray& array_)
{
	QString s;
	QByteArray b(array_.toHex());

	for (int i = 0; i < b.size(); i++)
		s.append(b.at(i));

	WRITE_TRACE(DBG_INFO, "%s: %s", name_, qPrintable(s));
}

} // namespace


///////////////////////////////////////////////////////////////////////////////
// struct Visitor

struct Visitor: boost::static_visitor<Prl::Expected<bool, int> >
{
	Visitor(const std::string& username_, const std::string& password_):
		m_username(username_), m_password(password_)
	{
	}

	result_type operator()(const Linux& engine_) const;
	result_type operator()(const Windows::Engine& engine_) const;

private:
	const std::string m_username;
	const std::string m_password;
};

Visitor::result_type Visitor::operator()(const Linux& engine_) const
{
	Prl::Expected<FILE*, int> s = engine_.pullShadow();
	if (s.isFailed())
		return s.error();

	std::string w;
	struct spwd *p = NULL;
	while (true)
	{
		p = fgetspent(s.value());
		if (NULL == p)
			break;
		if (0 == m_username.compare(p->sp_namp))
		{
			w = p->sp_pwdp;
			break;
		}
	}
	fclose(s.value());
	if (w.empty())
		return false;

	char* e = crypt(m_password.data(), w.data());
	if (NULL == e)
	{
		WRITE_TRACE(DBG_FATAL, "cannot crypt");
		return -1;
	}
	return 0 == w.compare(e);
}

Visitor::result_type Visitor::operator()(const Windows::Engine& engine_) const
{
	Prl::Expected<QByteArray, int> sysKey = engine_.getSyskey();
	if (sysKey.isFailed())
		return sysKey.error();

	Prl::Expected<Windows::SAM, int> s = engine_.getSAM();
	if (s.isFailed())
		return s.error();

	Prl::Expected<QByteArray, int> domAccF = s.value().getF();
	if (domAccF.isFailed())
		return domAccF.error();
	
	Prl::Expected<Windows::Account, int> v = s.value().getV(m_username);
	if (v.isFailed())
		return v.error();

	Windows::Magic m;
	QByteArray t = m.hash(m_password.data()), u;
	if (v.value().getPasswordSize() == 0)
		return t == u;

	Prl::Expected<QByteArray, int> samKey = m.getSamKey(domAccF.value(),
			sysKey.value(), v.value().getPasswordSize());
	if (samKey.isFailed())
		return v.error();

	QByteArray hash = m.decode(samKey.value(), v.value());
	u = m.decrypt(hash, v.value());

	printHex("Sys key", sysKey.value());
	printHex("Sam key", samKey.value());
	printHex("Hash", hash);
	printHex("Pw hash", t);
	printHex("NtPw hash", u);;

	return t == u;
}

///////////////////////////////////////////////////////////////////////////////
// struct Abstract

struct Abstract: boost::noncopyable
{
	Abstract();
	~Abstract();

	int operator()();
	int addImage(const std::string& path_);
	void setUsername(const std::string& value_)
	{
		m_username = value_;
	}
	void setPassword(const std::string& value_)
	{
		m_password = value_;
	}

private:
	guestfs_h* m_handle;
	std::string m_username;
	std::string m_password;
};

Abstract::Abstract(): m_handle(guestfs_create())
{
	if (NULL == m_handle)
		WRITE_TRACE(DBG_FATAL, "guestfs_create failed");
}

Abstract::~Abstract()
{
	guestfs_close(m_handle);
}

int Abstract::addImage(const std::string& path_)
{
	if (NULL == m_handle)
		return -1;

	if (-1 == guestfs_add_drive_opts(m_handle, path_.data(), GUESTFS_ADD_DRIVE_OPTS_READONLY, 1, -1))
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_add_drive_opts: %s", guestfs_last_error(m_handle));
		return -1;
	}
	return 0;
}

int Abstract::operator()()
{
	if (-1 == guestfs_launch(m_handle))
	{
		WRITE_TRACE(DBG_FATAL, "guestfs_launch: %s", guestfs_last_error(m_handle));
		return -1;
	}
	Facade f(m_handle);
	Prl::Expected<std::string, int> r = f.getOsRoot();
	if (r.isFailed())
		return r.error();

	Prl::Expected<void, int> m = f.mount(r.value());
	if (m.isFailed())
		return m.error();

	Prl::Expected<Facade::os_type, int> t = f.getOsType(r.value());
	if (t.isFailed())
		return t.error();

	Visitor::result_type x = boost::apply_visitor
		(Visitor(m_username, m_password), t.value());
	if (x.isFailed())
		return x.error();

	return !x.value();
}

///////////////////////////////////////////////////////////////////////////////
// struct Director

struct Director
{
	Director(int argc_, char** argv_);

	int operator()(Abstract& builder_);

private:
	po::command_line_parser m_parser;
	po::options_description m_description;
	po::positional_options_description m_positional;
};

Director::Director(int argc_, char** argv_): m_parser(argc_, argv_), m_description("Allowed options")
{
	m_description.add_options()
		("disk,d", po::value<std::vector<std::string> >(), "path to an image file")
		("verbose,v", "verbose")
		("username,u", po::value<std::string>(), "user name")
		("password,p", po::value<std::string>(), "password");
	m_positional.add("username", 1).add("password", 1);
	m_parser.options(m_description).positional(m_positional);
}

int Director::operator()(Abstract& builder_)
{
	do
	{
		po::variables_map m;
		try
		{
			po::store(m_parser.run(), m);
			po::notify(m);
		}
		catch(std::exception& e_)
		{
			std::cerr << e_.what() << std::endl;
			break;
		}
		if (0 == m.count("disk") || m.count("username") == 0)
			break;

		if (0 < m.count("verbose"))
			SetLogLevel(DBG_DEBUG);

		builder_.setUsername(m["username"].as<std::string>());
		BOOST_FOREACH(const std::string& d, m["disk"].as<std::vector<std::string> >())
		{
			int e = builder_.addImage(d);
			if (0 != e)
				return e;
		}
		if (0 < m.count("password"))
		{
			builder_.setPassword(m["password"].as<std::string>());
			return 0;
		}
		std::string w;
		std::cout << "password: ";
		if (std::getline(std::cin, w))
		{
			builder_.setPassword(w);
			return 0;
		}

	} while(false);

	std::cerr << m_description;
	return -1;
}

int main(int argc_, char** argv_)
{
	SetConsoleLogging(false);
	SetLogFileName(GetDefaultLogFilePath(), "vmauth.log");

	Abstract a;
	Director d(argc_, argv_);
	int e = d(a);
	if (0 != e)
		return e;
	  
	return a();
}

